# Copyright 2024 IQM Benchmarks developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module to initialize logger.

Should be used in every module to output information/warnings.
"""

import logging


# Configure the logger to log messages to the console
logging.basicConfig(level=logging.INFO, format="%(asctime)s - %(name)s - %(levelname)s - %(message)s")
# Create the logger
qcvv_logger = logging.getLogger(__name__)

# Adjust logging level for Qiskit logger
qiskit_logger = logging.getLogger("qiskit")
qiskit_logger.setLevel(logging.WARNING)



# Copyright 2024 IQM Benchmarks developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse

from iqm.benchmarks.randomized_benchmarking.interleaved_rb.interleaved_rb import *
from iqm.benchmarks.quantum_volume.quantum_volume import *
from iqm.benchmarks.entanglement.ghz import *


if __name__ == "__main__":
    parser = argparse.ArgumentParser("iqm.benchmark")
    parser.add_argument(
        "--backend",
        help="The type of backend (device) to be characterized",
        required=False,
        type=str,
        default="IQMFakeBackend",
    )
    args = parser.parse_args()

    if args.backend is not None and type(args.backend) is str and args.backend != "":
        backend = args.backend
    else:
        backend = "IQMFakeBackend"

    print(f"USING BACKEND: {backend}")

    from scheduled_experiments.adonis.weekly import *

    InterleavedRandomizedBenchmarking(backend, WEEKLY_IRB).run(backend)
    QuantumVolumeBenchmark(backend, WEEKLY_QV).run(backend)
    GHZBenchmark(backend, WEEKLY_GHZ).run(backend)



"""
Mirror Randomized Benchmarking.
"""

from copy import deepcopy
import random
from time import strftime
from typing import Any, Dict, List, Optional, Sequence, Tuple, Type, cast
import warnings

import numpy as np
from qiskit import QuantumCircuit, transpile
from qiskit.quantum_info import random_clifford, random_pauli
from qiskit_aer import Aer
from scipy.spatial.distance import hamming
import xarray as xr

from iqm.benchmarks import BenchmarkAnalysisResult, BenchmarkRunResult
from iqm.benchmarks.benchmark import BenchmarkConfigurationBase
from iqm.benchmarks.benchmark_definition import Benchmark, add_counts_to_dataset
from iqm.benchmarks.logging_config import qcvv_logger
from iqm.benchmarks.randomized_benchmarking.randomized_benchmarking_common import (
    exponential_rb,
    fit_decay_lmfit,
    lmfit_minimizer,
    plot_rb_decay,
    validate_irb_gate,
)
from iqm.benchmarks.utils import (
    get_iqm_backend,
    retrieve_all_counts,
    retrieve_all_job_metadata,
    submit_execute,
    timeit,
    xrvariable_to_counts,
)
from iqm.qiskit_iqm.iqm_backend import IQMBackendBase


def compute_polarizations(
    num_qubits: int,
    noisy_counts: List[Dict[str, int]],
    ideal_counts: List[Dict[str, int]],
    num_circ_samples: int,
    num_pauli_samples: int,
) -> List[float]:
    """Estimates the polarization for a list of noisy counts with respect to corresponding ideal counts
    The polarization here is a rescaling of the average fidelity, that corrects for few-qubit effects

    Arguments:
        num_qubits (int): the number of qubits being benchmarked
        noisy_counts (List[Dict[str, int]]): the list of counts coming from real execution
        ideal_counts (List[Dict[str, int]]): the list of counts coming from simulated, ideal execution
        num_circ_samples (int): the number circuit of samples used to estimate the polarization
        num_pauli_samples (int): the number of pauli samples per circuit sample used to estimate the polarization
    Returns:
        List[float]: the polarizations for each circuit sample of the given sequence length
    """
    ideal_counts_matrix = list_to_numcircuit_times_numpauli_matrix(ideal_counts, num_circ_samples, num_pauli_samples)
    noisy_counts_matrix = list_to_numcircuit_times_numpauli_matrix(noisy_counts, num_circ_samples, num_pauli_samples)

    polarizations: List = []
    for ideal_m, noisy_m in zip(ideal_counts_matrix, noisy_counts_matrix):
        polarization_pauli = []
        for ideal_pauli, noisy_pauli in zip(ideal_m, noisy_m):
            ideal_bitstrings = list(ideal_pauli.keys())
            noisy_bitstrings = list(noisy_pauli.keys())

            # Get the counts for each Hamming distance
            hamming_distances = {i: 0 for i in range(num_qubits + 1)}
            for s_n in noisy_bitstrings:
                for s_i in ideal_bitstrings:
                    k_index = int(hamming(list(s_n), list(s_i)) * num_qubits)
                    hamming_distances[k_index] += noisy_pauli[s_n]

            # Compute polarizations
            no_shots = sum(noisy_pauli.values())
            weighted_hamming = sum(((-1 / 2) ** k) * (hamming_distances[k] / no_shots) for k in range(num_qubits + 1))
            polarization_c = ((4**num_qubits) * weighted_hamming - 1) / (4**num_qubits - 1)
            polarization_pauli.append(polarization_c)

        # Average over the Pauli samples and add the circuit sample
        polarizations.append(np.mean(polarization_pauli))

    return polarizations


# TODO: Let edge_grab also admit a 1Q gate ensemble! Currently uniform Clifford by default # pylint: disable=fixme
# pylint: disable=too-many-branches
def edge_grab(
    qubit_set: List[int],
    n_layers: int,
    backend_arg: IQMBackendBase | str,
    density_2q_gates: float = 0.25,
    two_qubit_gate_ensemble: Optional[Dict[str, float]] = None,
) -> List[QuantumCircuit]:
    """Generate a list of random layers containing single-qubit Cliffords and two-qubit gates,
    sampled according to the edge-grab algorithm (see arXiv:2204.07568 [quant-ph]).

    Args:
        qubit_set (List[int]): The set of qubits of the backend.
        n_layers (int): The number of layers.
        backend_arg (IQMBackendBase | str): IQM backend.
        density_2q_gates (float): The expected density of 2Q gates in a circuit formed by subsequent application of layers
        two_qubit_gate_ensemble (Dict[str, float]): A dictionary with keys being str specifying 2Q gates, and values being corresponding probabilities
    Raises:
        ValueError: if the probabilities in the gate ensembles do not add up to unity.
    Returns:
        List[QuantumCircuit]: the list of gate layers, in the form of quantum circuits.
    """
    # Check the ensemble of 2Q gates, otherwise assign
    if two_qubit_gate_ensemble is None:
        two_qubit_gate_ensemble = cast(Dict[str, float], {"CZGate": 1.0})
    elif sum(two_qubit_gate_ensemble.values()) != 1.0:
        raise ValueError("The 2Q gate ensemble probabilities must sum to 1.0")

    # Validate 2Q gates and get circuits
    two_qubit_circuits = {}
    for k in two_qubit_gate_ensemble.keys():
        two_qubit_circuits[k] = validate_irb_gate(k, backend_arg, gate_params=None)
    # TODO: Admit parametrized 2Q gates! # pylint: disable=fixme

    # Check backend and retrieve if necessary
    if isinstance(backend_arg, str):
        backend = get_iqm_backend(backend_arg)
    else:
        backend = backend_arg

    # Definitions
    num_qubits = len(qubit_set)
    physical_to_virtual_map = {q: i for i, q in enumerate(qubit_set)}

    # Get the possible edges where to place 2Q gates given the backend connectivity
    twoq_edges = []
    for i, q0 in enumerate(qubit_set):
        for q1 in qubit_set[i + 1 :]:
            if (q0, q1) in list(backend.coupling_map):
                twoq_edges.append([q0, q1])
    twoq_edges = list(sorted(twoq_edges))

    # Generate the layers
    layer_list = []
    for _ in range(n_layers):
        # Pick edges at random and store them in a new list "edge_list"
        aux = deepcopy(twoq_edges)
        edge_list = []
        layer = QuantumCircuit(num_qubits)
        # Take (and remove) edges from "aux", then add to "edge_list"
        edge_qubits = []
        while aux:
            new_edge = random.choice(aux)
            edge_list.append(new_edge)
            edge_qubits = list(np.array(edge_list).flatten())
            # Removes all edges which include either of the qubits in new_edge
            aux = [e for e in aux if ((new_edge[0] not in e) and (new_edge[1] not in e))]

        # Define the probability for adding 2Q gates, given the input density
        if len(edge_list) != 0:
            prob_2qgate = num_qubits * density_2q_gates / len(edge_list)
        else:
            prob_2qgate = 0

        # Add gates in selected edges
        for e in edge_list:
            # Sample the 2Q gate
            two_qubit_gate = random.choices(
                list(two_qubit_gate_ensemble.keys()),
                weights=list(two_qubit_gate_ensemble.values()),
                k=1,
            )[0]

            # Pick whether to place the sampled 2Q gate according to the probability above
            is_gate_placed = random.choices(
                [True, False],
                weights=[prob_2qgate, 1 - prob_2qgate],
                k=1,
            )[0]

            if is_gate_placed:
                if two_qubit_gate == "clifford":
                    layer.compose(
                        random_clifford(2).to_instruction(),
                        qubits=[
                            physical_to_virtual_map[e[0]],
                            physical_to_virtual_map[e[1]],
                        ],
                        inplace=True,
                    )
                else:
                    layer.append(
                        two_qubit_circuits[two_qubit_gate],
                        [
                            physical_to_virtual_map[e[0]],
                            physical_to_virtual_map[e[1]],
                        ],
                    )
            else:
                layer.compose(
                    random_clifford(1).to_instruction(),
                    qubits=[physical_to_virtual_map[e[0]]],
                    inplace=True,
                )
                layer.compose(
                    random_clifford(1).to_instruction(),
                    qubits=[physical_to_virtual_map[e[1]]],
                    inplace=True,
                )

        # Add 1Q gates in remaining qubits
        remaining_qubits = [q for q in qubit_set if q not in edge_qubits]
        while remaining_qubits:
            for q in remaining_qubits:
                layer.compose(
                    random_clifford(1).to_instruction(),
                    qubits=[physical_to_virtual_map[q]],
                    inplace=True,
                )
                remaining_qubits.remove(q)

        layer_list.append(layer)

    return layer_list


def generate_pauli_dressed_mrb_circuits(
    qubits: List[int],
    pauli_samples_per_circ: int,
    depth: int,
    backend_arg: IQMBackendBase | str,
    density_2q_gates: float = 0.25,
    two_qubit_gate_ensemble: Optional[Dict[str, float]] = None,
    qiskit_optim_level: int = 1,
    routing_method: str = "basic",
) -> Dict[str, List[QuantumCircuit]]:
    """Samples a mirror circuit and generates samples of "Pauli-dressed" circuits,
        where for each circuit, random Pauli layers are interleaved between each layer of the circuit

    Args:
        qubits (List[int]): the qubits of the backend
        pauli_samples_per_circ (int): the number of pauli samples per circuit
        depth (int): the depth (number of canonical layers) of the circuit
        backend_arg (IQMBackendBase | str): the backend
        density_2q_gates (float): the expected density of 2Q gates
        two_qubit_gate_ensemble (Optional[Dict[str, float]]):
        qiskit_optim_level (int):
        routing_method (str):
    Returns:

    """
    num_qubits = len(qubits)

    # Sample the layers using edge grab sampler - different samplers may be conditionally chosen here in the future
    cycle_layers = edge_grab(qubits, depth, backend_arg, density_2q_gates, two_qubit_gate_ensemble)

    # Sample the edge (initial/final) random Single-qubit Clifford layer
    clifford_layer = [random_clifford(1) for _ in range(num_qubits)]

    # Initialize the list of circuits
    all_circuits = {}
    pauli_dressed_circuits_untranspiled: List[QuantumCircuit] = []
    pauli_dressed_circuits_transpiled: List[QuantumCircuit] = []

    for _ in range(pauli_samples_per_circ):
        # Initialize the quantum circuit object
        circ = QuantumCircuit(num_qubits)
        # Sample all the random Paulis
        paulis = [random_pauli(num_qubits) for _ in range(depth + 1)]

        # Add the edge product of Cliffords
        for i in range(num_qubits):
            circ.compose(clifford_layer[i].to_instruction(), qubits=[i], inplace=True)
        circ.barrier()

        # Add the cycle layers
        for k in range(depth):
            circ.compose(
                paulis[k].to_instruction(),
                qubits=list(range(num_qubits)),
                inplace=True,
            )
            circ.barrier()
            circ.compose(cycle_layers[k], inplace=True)
            circ.barrier()

        # Apply middle Pauli
        circ.compose(
            paulis[depth].to_instruction(),
            qubits=list(range(num_qubits)),
            inplace=True,
        )
        circ.barrier()

        # Add the mirror layers
        for k in range(depth):
            circ.compose(cycle_layers[depth - k - 1].inverse(), inplace=True)
            circ.barrier()
            circ.compose(
                paulis[depth - k - 1].to_instruction(),
                qubits=list(range(num_qubits)),
                inplace=True,
            )
            circ.barrier()

        # Add the inverse edge product of Cliffords
        for i in range(num_qubits):
            circ.compose(clifford_layer[i].to_instruction().inverse(), qubits=[i], inplace=True)

        # Add measurements
        circ.measure_all()

        # Transpile to backend - no optimize SQG should be used!
        if isinstance(backend_arg, str):
            retrieved_backend = get_iqm_backend(backend_arg)
        else:
            assert isinstance(backend_arg, IQMBackendBase)
            retrieved_backend = backend_arg

        circ_transpiled = transpile(
            circ,
            backend=retrieved_backend,
            initial_layout=qubits,
            optimization_level=qiskit_optim_level,
            routing_method=routing_method,
        )

        pauli_dressed_circuits_untranspiled.append(circ)
        pauli_dressed_circuits_transpiled.append(circ_transpiled)

    # Store the circuit
    all_circuits.update(
        {
            "untranspiled": pauli_dressed_circuits_untranspiled,
            "transpiled": pauli_dressed_circuits_transpiled,
        }
    )

    return all_circuits


@timeit
def generate_fixed_depth_mrb_circuits(
    qubits: List[int],
    circ_samples: int,
    pauli_samples_per_circ: int,
    depth: int,
    backend_arg: IQMBackendBase | str,
    density_2q_gates: float = 0.25,
    two_qubit_gate_ensemble: Optional[Dict[str, float]] = None,
    qiskit_optim_level: int = 1,
    routing_method: str = "basic",
) -> Dict[int, Dict[str, List[QuantumCircuit]]]:
    """Generates a dictionary MRB circuits at fixed depth, indexed by sample number

    Args:
        qubits (List[int]): A list of integers specifying physical qubit labels
        circ_samples (int): The number of sets of Pauli-dressed circuit samples
        pauli_samples_per_circ (int): the number of pauli samples per circuit
        depth (int): the depth (number of canonical layers) of the circuits
        backend_arg (IQMBackendBase | str): the backend
        density_2q_gates (float):
        two_qubit_gate_ensemble (Optional[Dict[str, float]]):
        qiskit_optim_level (int):
        routing_method (str):
    Returns:
        A dictionary of lists of Pauli-dressed quantum circuits corresponding to the circuit sample index
    """

    circuits = {}  # The dict with a Dict of random mirror circuits
    for p_sample in range(circ_samples):
        circuits[p_sample] = generate_pauli_dressed_mrb_circuits(
            qubits,
            pauli_samples_per_circ,
            depth,
            backend_arg,
            density_2q_gates,
            two_qubit_gate_ensemble,
            qiskit_optim_level,
            routing_method,
        )

    return circuits


def list_to_numcircuit_times_numpauli_matrix(
    input_list: List[Any], num_circ_samples: int, num_pauli_samples: int
) -> List[List[Any]]:
    """Convert a flat list to a matrix of shape (num_circ_samples, num_pauli_samples).
    Args:
        input_list (List[Any]): the input flat list
        num_circ_samples (int): the number of sets of Pauli-dressed circuit samples
        num_pauli_samples (int): the number of Pauli samples per circuit
    Raises:
        ValueError: Length of passed list is not (num_circ_samples * num_pauli_samples).
    Returns:
        List[List[Any]]: the matrix
    """
    if len(input_list) != num_circ_samples * num_pauli_samples:
        raise ValueError(
            f"Length of passed list {len(input_list)} is not"
            f" (num_circ_samples * num_pauli_samples) = {num_circ_samples * num_pauli_samples}"
        )

    return np.reshape(input_list, (num_circ_samples, num_pauli_samples)).tolist()


# pylint: disable=too-many-statements
def mrb_analysis(run: BenchmarkRunResult) -> BenchmarkAnalysisResult:
    """Analysis function for a MRB experiment

    Args:
        run (RunResult): A MRB experiment run for which analysis result is created
    Returns:
        AnalysisResult corresponding to MRB
    """
    plots = {}
    observations = {}
    dataset = run.dataset.copy(deep=True)

    shots = dataset.attrs["shots"]
    num_circuit_samples = dataset.attrs["num_circuit_samples"]
    num_pauli_samples = dataset.attrs["num_pauli_samples"]

    density_2q_gates = dataset.attrs["density_2q_gates"]
    two_qubit_gate_ensemble = dataset.attrs["two_qubit_gate_ensemble"]

    max_gates_per_batch = dataset.attrs["max_gates_per_batch"]

    # Analyze the results for each qubit layout of the experiment dataset
    qubits_array = dataset.attrs["qubits_array"]
    depths_array = dataset.attrs["depths_array"]

    assigned_mrb_depths = {}
    if len(qubits_array) != len(depths_array):
        # If user did not specify a list of depth for each list of qubits, assign the first
        # If the len is not one, the input was incorrect
        if len(depths_array) != 1:
            qcvv_logger.info(
                f"The amount of qubit layouts ({len(qubits_array)}) is not the same "
                f"as the amount of depth configurations ({len(depths_array)}):\n\tWill assign to all the first "
                f"configuration: {depths_array[0]} !"
            )
        assigned_mrb_depths = {str(q): [2 * m for m in depths_array[0]] for q in qubits_array}
    else:
        assigned_mrb_depths = {str(qubits_array[i]): [2 * m for m in depths_array[i]] for i in range(len(depths_array))}

    transpiled_circuits = dataset.attrs["transpiled_circuits"]
    simulator = Aer.get_backend("qasm_simulator")

    all_noisy_counts: Dict[str, Dict[int, List[Dict[str, int]]]] = {}
    all_noiseless_counts: Dict[str, Dict[int, List[Dict[str, int]]]] = {}
    time_retrieve_noiseless: Dict[str, Dict[int, float]] = {}
    # Need to loop over each set of qubits, and within, over each depth
    for qubits_idx, qubits in enumerate(qubits_array):
        polarizations = {}
        num_qubits = len(qubits)
        all_noisy_counts[str(qubits)] = {}
        all_noiseless_counts[str(qubits)] = {}
        time_retrieve_noiseless[str(qubits)] = {}
        qcvv_logger.info(f"Post-processing MRB for qubits {qubits}")
        for depth in assigned_mrb_depths[str(qubits)]:
            # Retrieve counts
            identifier = f"qubits_{str(qubits)}_depth_{str(depth)}"
            all_noisy_counts[str(qubits)][depth] = xrvariable_to_counts(
                dataset, identifier, num_circuit_samples * num_pauli_samples
            )

            qcvv_logger.info(f"Depth {depth}")
            # Execute the quantum circuits on the simulated, ideal backend
            # pylint: disable=unbalanced-tuple-unpacking
            all_noiseless_jobs, _ = submit_execute(
                {tuple(qubits): transpiled_circuits[str(qubits)][str(depth)]},
                simulator,
                shots,
                calset_id=None,
                max_gates_per_batch=max_gates_per_batch,
            )

            # Retrieve counts
            all_noiseless_counts[str(qubits)][depth], time_retrieve_noiseless[str(qubits)][depth] = retrieve_all_counts(
                all_noiseless_jobs
            )

            # Compute polarizations for the current depth
            polarizations[depth] = compute_polarizations(
                num_qubits,
                all_noisy_counts[str(qubits)][depth],
                all_noiseless_counts[str(qubits)][depth],
                num_circuit_samples,
                num_pauli_samples,
            )

        # Fit decay and extract parameters
        list_of_polarizations = list(polarizations.values())
        fit_data, fit_parameters = fit_decay_lmfit(exponential_rb, qubits, list_of_polarizations, "mrb")
        rb_fit_results = lmfit_minimizer(fit_parameters, fit_data, assigned_mrb_depths[str(qubits)], exponential_rb)

        average_polarizations = {d: np.mean(polarizations[d]) for d in assigned_mrb_depths[str(qubits)]}
        stddevs_from_mean = {
            d: np.std(polarizations[d]) / np.sqrt(num_circuit_samples * num_pauli_samples)
            for d in assigned_mrb_depths[str(qubits)]
        }
        popt = {
            "amplitude": rb_fit_results.params["amplitude_1"],
            "offset": rb_fit_results.params["offset_1"],
            "decay_rate": rb_fit_results.params["p_mrb"],
        }
        fidelity = rb_fit_results.params["fidelity_mrb"]

        processed_results = {
            "avg_gate_fidelity": {"value": fidelity.value, "uncertainty": fidelity.stderr},
        }

        dataset.attrs[qubits_idx].update(
            {
                "decay_rate": {"value": popt["decay_rate"].value, "uncertainty": popt["decay_rate"].stderr},
                "fit_amplitude": {"value": popt["amplitude"].value, "uncertainty": popt["amplitude"].stderr},
                "fit_offset": {"value": popt["offset"].value, "uncertainty": popt["offset"].stderr},
                "polarizations": polarizations,
                "avg_polarization_nominal_values": average_polarizations,
                "avg_polatization_stderr": stddevs_from_mean,
                "fitting_method": str(rb_fit_results.method),
                "num_function_evals": int(rb_fit_results.nfev),
                "data_points": int(rb_fit_results.ndata),
                "num_variables": int(rb_fit_results.nvarys),
                "chi_square": float(rb_fit_results.chisqr),
                "reduced_chi_square": float(rb_fit_results.redchi),
                "Akaike_info_crit": float(rb_fit_results.aic),
                "Bayesian_info_crit": float(rb_fit_results.bic),
            }
        )

        # Update observations
        observations.update({qubits_idx: processed_results})

        # Generate plots
        fig_name, fig = plot_rb_decay(
            "mrb",
            [qubits],
            dataset,
            observations,
            mrb_2q_density=density_2q_gates,
            mrb_2q_ensemble=two_qubit_gate_ensemble,
        )
        plots[fig_name] = fig

    # Generate the combined plot
    fig_name, fig = plot_rb_decay(
        "mrb",
        qubits_array,
        dataset,
        observations,
        mrb_2q_density=density_2q_gates,
        mrb_2q_ensemble=two_qubit_gate_ensemble,
    )
    plots[fig_name] = fig

    return BenchmarkAnalysisResult(dataset=dataset, plots=plots, observations=observations)


class MirrorRandomizedBenchmarking(Benchmark):
    """
    Mirror RB estimates the fidelity of ensembles of n-qubit layers
    """

    analysis_function = staticmethod(mrb_analysis)

    name: str = "mirror_rb"

    def __init__(self, backend_arg: IQMBackendBase | str, configuration: "MirrorRBConfiguration"):
        """Construct the MirrorRandomizedBenchmarking class

        Args:
            backend_arg (IQMBackendBase | str): _description_
            configuration (MirrorRBConfiguration): _description_
        """
        super().__init__(backend_arg, configuration)

        # EXPERIMENT
        self.backend_configuration_name = backend_arg if isinstance(backend_arg, str) else backend_arg.name

        self.qubits_array = configuration.qubits_array
        self.depths_array = configuration.depths_array
        self.num_circuit_samples = configuration.num_circuit_samples
        self.num_pauli_samples = configuration.num_pauli_samples
        self.two_qubit_gate_ensemble = configuration.two_qubit_gate_ensemble
        self.density_2q_gates = configuration.density_2q_gates

        self.qiskit_optim_level = configuration.qiskit_optim_level

        self.simulator = Aer.get_backend("qasm_simulator")

        self.session_timestamp = strftime("%Y%m%d-%H%M%S")
        self.execution_timestamp = ""

    def add_all_meta_to_dataset(self, dataset: xr.Dataset):
        """Adds all configuration metadata and circuits to the dataset variable

        Args:
            dataset (xr.Dataset): The xarray dataset
        """
        dataset.attrs["session_timestamp"] = self.session_timestamp
        dataset.attrs["execution_timestamp"] = self.execution_timestamp
        dataset.attrs["backend_configuration_name"] = self.backend_configuration_name
        dataset.attrs["backend_name"] = self.backend.name

        for key, value in self.configuration:
            if key == "benchmark":  # Avoid saving the class object
                dataset.attrs[key] = value.name
            else:
                dataset.attrs[key] = value
        # Defined outside configuration - if any

    @timeit
    def add_all_circuits_to_dataset(self, dataset: xr.Dataset):
        """Adds all generated circuits during execution to the dataset variable
        Args:
            dataset (xr.Dataset):  The xarray dataset
        """
        qcvv_logger.info(f"Adding all circuits to the dataset")
        for key, circuit in zip(
            ["transpiled_circuits", "untranspiled_circuits"], [self.transpiled_circuits, self.untranspiled_circuits]
        ):
            dictionary = {}
            for outer_key, outer_value in circuit.items():
                dictionary[str(outer_key)] = {
                    str(inner_key): inner_values for inner_key, inner_values in outer_value.items()
                }
            dataset.attrs[key] = dictionary

    def submit_single_mrb_job(
        self,
        backend_arg: IQMBackendBase,
        qubits: Sequence[int],
        depth: int,
        sorted_transpiled_circuit_dicts: Dict[Tuple[int, ...], List[QuantumCircuit]],
    ) -> Dict[str, Any]:
        """
            Submit fixed-depth MRB jobs for execution in the specified IQMBackend
        Args:
            backend_arg (IQMBackendBase): the IQM backend to submit the job
            qubits (Sequence[int]): the qubits to identify the submitted job
            depth (int): the depth (number of canonical layers) of the circuits to identify the submitted job
            sorted_transpiled_circuit_dicts (Dict[str, List[QuantumCircuit]]): A dictionary containing all MRB circuits
        Returns:
            Dict with qubit layout, submitted job objects, type (vanilla/DD) and submission time
        """
        # Submit
        # Send to execute on backend
        execution_jobs, time_submit = submit_execute(
            sorted_transpiled_circuit_dicts,
            backend_arg,
            self.shots,
            self.calset_id,
            max_gates_per_batch=self.max_gates_per_batch,
        )
        mrb_submit_results = {
            "qubits": qubits,
            "depth": depth,
            "jobs": execution_jobs,
            "time_submit": time_submit,
        }
        return mrb_submit_results

    def execute(self, backend: IQMBackendBase) -> xr.Dataset:
        """Executes the benchmark"""

        self.execution_timestamp = strftime("%Y%m%d-%H%M%S")

        dataset = xr.Dataset()
        self.add_all_meta_to_dataset(dataset)

        # Submit jobs for all qubit layouts
        all_mrb_jobs: List[Dict[str, Any]] = []
        time_circuit_generation: Dict[str, float] = {}

        # Initialize the variable to contain the circuits for each layout
        self.untranspiled_circuits: Dict[str, Dict[int | str, List[QuantumCircuit]]] = {}
        self.transpiled_circuits: Dict[str, Dict[int | str, List[QuantumCircuit]]] = {}

        # The depths should be assigned to each set of qubits!
        # The real final MRB depths are twice the originally specified, must be taken into account here!
        assigned_mrb_depths = {}
        if len(self.qubits_array) != len(self.depths_array):
            # If user did not specify a list of depth for each list of qubits, assign the first
            # If the len is not one, the input was incorrect
            if len(self.depths_array) != 1:
                warnings.warn(
                    f"The amount of qubit layouts ({len(self.qubits_array)}) is not the same "
                    f"as the amount of depth configurations ({len(self.depths_array)}):\n\tWill assign to all the first "
                    f"configuration: {self.depths_array[0]} !"
                )
            assigned_mrb_depths = {str(q): [2 * m for m in self.depths_array[0]] for q in self.qubits_array}
        else:
            assigned_mrb_depths = {
                str(self.qubits_array[i]): [2 * m for m in self.depths_array[i]] for i in range(len(self.depths_array))
            }

        # Auxiliary dict from str(qubits) to indices
        qubit_idx: Dict[str, Any] = {}
        for qubits_idx, qubits in enumerate(self.qubits_array):
            qubit_idx[str(qubits)] = qubits_idx
            self.untranspiled_circuits[str(qubits)] = {}
            self.transpiled_circuits[str(qubits)] = {}

            qcvv_logger.info(
                f"Executing MRB on qubits {qubits}."
                f" Will generate and submit all {self.num_circuit_samples}x{self.num_pauli_samples} MRB circuits"
                f" for each depth {assigned_mrb_depths[str(qubits)]}"
            )
            mrb_circuits = {}
            mrb_transpiled_circuits_lists: Dict[int, List[QuantumCircuit]] = {}
            mrb_untranspiled_circuits_lists: Dict[int, List[QuantumCircuit]] = {}
            time_circuit_generation[str(qubits)] = 0
            for depth in assigned_mrb_depths[str(qubits)]:
                qcvv_logger.info(f"Depth {depth}")
                mrb_circuits[depth], elapsed_time = generate_fixed_depth_mrb_circuits(
                    qubits,
                    self.num_circuit_samples,
                    self.num_pauli_samples,
                    int(depth / 2),
                    backend,
                    self.density_2q_gates,
                    self.two_qubit_gate_ensemble,
                    self.qiskit_optim_level,
                    self.routing_method,
                )
                time_circuit_generation[str(qubits)] += elapsed_time

                # Generated circuits at fixed depth are (dict) indexed by Pauli sample number, turn into List
                mrb_transpiled_circuits_lists[depth] = []
                mrb_untranspiled_circuits_lists[depth] = []
                for c_s in range(self.num_circuit_samples):
                    mrb_transpiled_circuits_lists[depth].extend(mrb_circuits[depth][c_s]["transpiled"])
                for c_s in range(self.num_circuit_samples):
                    mrb_untranspiled_circuits_lists[depth].extend(mrb_circuits[depth][c_s]["untranspiled"])

                # Submit
                sorted_transpiled_qc_list = {tuple(qubits): mrb_transpiled_circuits_lists[depth]}
                all_mrb_jobs.append(self.submit_single_mrb_job(backend, qubits, depth, sorted_transpiled_qc_list))
                qcvv_logger.info(f"Job for layout {qubits} & depth {depth} submitted successfully!")

            self.untranspiled_circuits[str(qubits)] = {
                str(d): mrb_untranspiled_circuits_lists[d] for d in assigned_mrb_depths[str(qubits)]
            }
            self.transpiled_circuits[str(qubits)] = {
                str(d): mrb_transpiled_circuits_lists[d] for d in assigned_mrb_depths[str(qubits)]
            }

            dataset.attrs[qubits_idx] = {"qubits": qubits}

        # Retrieve counts of jobs for all qubit layouts
        all_job_metadata = {}
        for job_dict in all_mrb_jobs:
            qubits = job_dict["qubits"]
            depth = job_dict["depth"]
            # Retrieve counts
            execution_results, time_retrieve = retrieve_all_counts(
                job_dict["jobs"], f"qubits_{str(qubits)}_depth_{str(depth)}"
            )
            # Retrieve all job meta data
            all_job_metadata = retrieve_all_job_metadata(job_dict["jobs"])
            # Export all to dataset
            dataset.attrs[qubit_idx[str(qubits)]].update(
                {
                    f"depth_{str(depth)}": {
                        "time_circuit_generation": time_circuit_generation[str(qubits)],
                        "time_submit": job_dict["time_submit"],
                        "time_retrieve": time_retrieve,
                        "all_job_metadata": all_job_metadata,
                    },
                }
            )

            qcvv_logger.info(f"Adding counts of qubits {qubits} and depth {depth} run to the dataset")
            dataset, _ = add_counts_to_dataset(execution_results, f"qubits_{str(qubits)}_depth_{str(depth)}", dataset)

        self.add_all_circuits_to_dataset(dataset)

        qcvv_logger.info(f"MRB experiment execution concluded !")

        return dataset


class MirrorRBConfiguration(BenchmarkConfigurationBase):
    """Mirror RB configuration.

    Attributes:
        benchmark (Type[Benchmark]): MirrorRandomizedBenchmarking.
        qubits_array (Sequence[Sequence[int]]): The array of physical qubits in which to execute MRB.
        depths_array (Sequence[Sequence[int]]): The array of physical depths in which to execute MRB for a corresponding qubit list.
                            * If len is the same as that of qubits_array, each Sequence[int] corresponds to the depths for the corresponding layout of qubits.
                            * If len is different from that of qubits_array, assigns the first Sequence[int].
        num_circuit_samples (int): The number of random-layer mirror circuits to generate.
        num_pauli_samples (int): The number of random Pauli layers to interleave per mirror circuit.
        shots (int): The number of measurement shots to execute per circuit.
        qiskit_optim_level (int): The Qiskit-level of optimization to use in transpilation.
                            * Default is 1.
        routing_method (Literal["basic", "lookahead", "stochastic", "sabre", "none"]): The routing method to use in transpilation.
                            * Default is "sabre".
        two_qubit_gate_ensemble (Dict[str, float]): The two-qubit gate ensemble to use in the random mirror circuits.
                            * Keys correspond to str names of qiskit circuit library gates, e.g., "CZGate" or "CXGate".
                            * Values correspond to the probability for the respective gate to be sampled.
                            * Default is {"CZGate": 1.0}.
        density_2q_gates (float): The expected density of 2-qubit gates in the final circuits.
                            * Default is 0.25.
    """

    benchmark: Type[Benchmark] = MirrorRandomizedBenchmarking
    qubits_array: Sequence[Sequence[int]]
    depths_array: Sequence[Sequence[int]]
    num_circuit_samples: int
    num_pauli_samples: int
    qiskit_optim_level: int = 1
    two_qubit_gate_ensemble: Dict[str, float] = {
        "CZGate": 1.0,
    }
    density_2q_gates: float = 0.25



"""Tests for compressive GST benchmark"""

from iqm.benchmarks.compressive_gst.compressive_gst import GSTConfiguration, CompressiveGST
from unittest.mock import patch

backend = "iqmfakeapollo"


class TestGST:
    @patch('matplotlib.pyplot.figure')
    def test_1q(self, mock_fig):
        minimal_1Q_config = GSTConfiguration(
            qubit_layouts=[[5], [0]],
            gate_set="1QXYI",
            num_circuits=10,
            shots=10,
            rank=4,
            bootstrap_samples=2,
            max_iterations=[1, 1]
        )
        benchmark = CompressiveGST(backend, minimal_1Q_config)
        benchmark.run()
        result = benchmark.analyze()
        print(result.observations)
        mock_fig.assert_called()

    @patch('matplotlib.pyplot.figure')
    def test_2q(self, mock_fig):
        minimal_2Q_GST = GSTConfiguration(
            qubit_layouts=[[0, 1]],
            gate_set="2QXYCZ_extended",
            num_circuits=4,
            shots=10,
            rank=1,
            bootstrap_samples=0,
            max_iterations=[1, 1]
        )
        benchmark = CompressiveGST(backend, minimal_2Q_GST)
        benchmark.run()
        result = benchmark.analyze()
        mock_fig.assert_called()



# Copyright 2024 IQM Benchmarks developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
CLOPS benchmark
"""

from datetime import datetime
from math import floor, pi
from time import perf_counter, strftime
from typing import Any, Dict, List, Sequence, Tuple, Type

import matplotlib as mpl
from matplotlib.figure import Figure
import matplotlib.pyplot as plt
import numpy as np
from qiskit import QuantumCircuit
from qiskit.circuit import ParameterVector
import xarray as xr

from iqm.benchmarks import Benchmark
from iqm.benchmarks.benchmark import BenchmarkConfigurationBase
from iqm.benchmarks.benchmark_definition import BenchmarkAnalysisResult, BenchmarkRunResult
from iqm.benchmarks.logging_config import qcvv_logger
from iqm.benchmarks.utils import (
    count_2q_layers,
    count_native_gates,
    perform_backend_transpilation,
    retrieve_all_counts,
    retrieve_all_job_metadata,
    set_coupling_map,
    sort_batches_by_final_layout,
    submit_execute,
    timeit,
)
from iqm.qiskit_iqm.iqm_backend import IQMBackendBase
from iqm.qiskit_iqm.iqm_transpilation import optimize_single_qubit_gates


def plot_times(clops_data: xr.Dataset, observations: Dict[int, Dict[str, Dict[str, float]]]) -> Tuple[str, Figure]:
    """Generate a figure representing the different elapsed times in the CLOPS experiment.

    Args:
        clops_data (xr.Dataset): The dataset including elapsed time data from the CLOPS experiment
        observations (Dict[int, Dict[str, Dict[str, float]]]): The observations from the analysis of the CLOPS experiment.
    Returns:
        str: the name of the figure.
        Figure: the figure.
    """
    # Define the keys for different categories of times
    job_keys = ["submit_total", "compile_total", "execution_total"]
    user_keys = ["user_submit_total", "user_retrieve_total", "assign_parameters_total"]
    total_keys = ["job_total"]

    # Define variables for dataset values
    qubits = clops_data.attrs["qubits"]
    num_qubits = clops_data.attrs["num_qubits"]
    backend_name = clops_data.attrs["backend_name"]
    execution_timestamp = clops_data.attrs["execution_timestamp"]
    clops_time = float(clops_data.attrs["clops_time"])

    all_data = {}
    all_data.update(clops_data.attrs)
    all_data.update(observations[1])

    # Define colors
    cmap = mpl.pyplot.get_cmap("winter")
    colors = [cmap(i) for i in np.linspace(0, 1, len(job_keys) + len(user_keys) + len(total_keys) + 1)]

    # Plotting parameters
    fontsize = 6
    sep = 1
    barsize = 1
    alpha = 0.4

    fig, ax1 = plt.subplots()
    ax2 = ax1.twinx()

    fig_name = f"{num_qubits}_qubits_{tuple(qubits)}"

    # Plot total CLOPS time
    x_t = ax1.bar(4 * sep, clops_time, barsize, zorder=0, label="clops time", color=(colors[-1], alpha), edgecolor="k")
    ax1.bar_label(x_t, fmt=f"clops time: {clops_time:.2f}", fontsize=fontsize)

    # Plot user keys
    for i, (key, cumulative_value) in enumerate(zip(user_keys, np.cumsum([all_data[k] for k in user_keys]))):
        x = ax1.bar(
            3 * sep,
            cumulative_value,
            barsize,
            zorder=1 - i / 10,
            label=key,
            color=(colors[len(job_keys) + len(total_keys) + i], alpha),
            edgecolor="k",
        )
        ax1.bar_label(x, fmt=f"{key.replace('_total', ' ').replace('_', ' ')}: {all_data[key]:.2f}", fontsize=fontsize)

    # Plot total keys
    for i, (key, cumulative_value) in enumerate(zip(total_keys, np.cumsum([all_data[k] for k in total_keys]))):
        x = ax1.bar(
            2 * sep,
            cumulative_value,
            barsize,
            zorder=1 - i / 10,
            label=key,
            color=(colors[len(job_keys) + i], alpha),
            edgecolor="k",
        )
        ax1.bar_label(x, fmt=f"{key.replace('_total', ' ')}: {all_data[key]:.2f}", fontsize=fontsize)

    # Plot job keys
    for i, (key, cumulative_value) in enumerate(zip(job_keys, np.cumsum([all_data[k] for k in job_keys]))):
        x = ax1.bar(
            sep, cumulative_value, barsize, zorder=2 - i / 10, label=key, color=(colors[i], alpha), edgecolor="k"
        )
        ax1.bar_label(x, label_type="edge", fmt=f"{key.replace('_total', ' ')}: {all_data[key]:.2f}", fontsize=fontsize)

    # Remove top and bottom spines
    ax1.spines["top"].set_visible(False)
    ax1.spines["bottom"].set_visible(False)
    ax2.spines["top"].set_visible(False)
    ax2.spines["bottom"].set_visible(False)

    # Set axis labels and limits
    ax1.set_ylabel("Total CLOPS time (seconds)")
    ax2.set_ylabel("Total CLOPS time (%)")
    ax1.set_ylim(-0.2, clops_time + 1)
    ax2.set_ylim(-0.2, 100)

    # Set x-ticks and labels
    time_types = ["Remote (components)", "Remote (total)", "Wall-time (components)", "Wall-time (total)"]
    ax1.set_xticks([i * sep + 1 for i in range(4)], time_types, fontsize=fontsize)

    # Set plot title
    if all_data["clops_h"]["value"] == 0:
        plt.title(
            f"Total execution times for CLOPS experiment\n"
            f"{backend_name}, {execution_timestamp}\n"
            f"CLOPS_v {all_data['clops_v']['value']} on qubits {qubits}\n"
        )
    else:
        plt.title(
            f"Total execution times for CLOPS experiment\n"
            f"{backend_name}, {execution_timestamp}\n"
            f"CLOPS_v {all_data['clops_v']['value']} / CLOPS_h {all_data['clops_h']['value']} on qubits {qubits}\n"
        )

    plt.gcf().set_dpi(250)
    plt.close()

    return fig_name, fig


def retrieve_clops_elapsed_times(job_meta: Dict[str, Dict[str, Any]]) -> Dict[str, float]:
    """Retrieve the elapsed times from the CLOPS job metadata

    Args:
        job_meta (Dict[Dict[str, Any]]): A dictionary of CLOPS jobs metadata
    Returns:
        Dict[str, float]: A dictionary of elapsed times of all CLOPS jobs
    """
    all_job_elapsed: Dict = {}
    totals_keys: Dict = {}
    for u_index, update in enumerate(job_meta):
        all_job_elapsed[update] = {}
        for b_index, batch in enumerate(job_meta[update].keys()):
            # ["timestamps"] might be empty if backend is a simulator
            if job_meta[update][batch]["timestamps"] is not None:
                x = job_meta[update][batch]["timestamps"]
                job_time_format = "%Y-%m-%dT%H:%M:%S.%f%z"  # Is it possible to extract this automatically?
                compile_f = datetime.strptime(x["compile_end"], job_time_format)
                compile_i = datetime.strptime(x["compile_start"], job_time_format)
                submit_f = datetime.strptime(x["submit_end"], job_time_format)
                submit_i = datetime.strptime(x["submit_start"], job_time_format)
                execution_f = datetime.strptime(x["execution_end"], job_time_format)
                execution_i = datetime.strptime(x["execution_start"], job_time_format)
                job_f = datetime.strptime(x["job_end"], job_time_format)
                job_i = datetime.strptime(x["job_start"], job_time_format)

                all_job_elapsed[update][batch] = {
                    "job_total": job_f - job_i,
                    "compile_total": compile_f - compile_i,
                    "submit_total": submit_f - submit_i,
                    "execution_total": execution_f - execution_i,
                }

                # Save the keys, will be needed later
                totals_keys = all_job_elapsed[update][batch].keys()

                # Update metadata file - need to turn values ("timedelta" type) into str
                job_meta["update_" + str(u_index + 1)]["batch_job_" + str(b_index + 1)].update(
                    {k: str(all_job_elapsed[update][batch][k]) for k in totals_keys}
                )

    # Add up totals
    overall_elapsed: Dict = {}
    # If backend is a simulator
    if not totals_keys:
        return overall_elapsed

    overall_elapsed = {k: 0 for k in totals_keys}
    overall_elapsed_per_update: Dict = {}
    for update in job_meta:
        overall_elapsed_per_update[update] = {k: 0 for k in totals_keys}
        for batch in job_meta[update].keys():
            for k in totals_keys:
                overall_elapsed_per_update[update][k] += all_job_elapsed[update][batch][k].total_seconds()
        for k in totals_keys:
            overall_elapsed[k] += overall_elapsed_per_update[update][k]

    return overall_elapsed


def clops_analysis(run: BenchmarkRunResult) -> BenchmarkAnalysisResult:
    """Analysis function for a CLOPS (v or h) experiment

    Args:
        run (RunResult): A CLOPS experiment run for which analysis result is created
    Returns:
        AnalysisResult corresponding to CLOPS
    """
    plots: Dict[str, Any] = {}
    observations = {}
    dataset = run.dataset

    # Retrieve dataset values
    # backend_name = dataset.attrs["backend_configuration_name"]
    num_circuits = dataset.attrs["num_circuits"]
    num_updates = dataset.attrs["num_updates"]
    num_shots = dataset.attrs["num_shots"]
    depth = dataset.attrs["depth"]

    clops_h_bool = dataset.attrs["clops_h_bool"]

    all_job_meta = dataset.attrs["job_meta_per_update"]

    clops_time = dataset.attrs["clops_time"]
    all_times_parameter_assign = dataset.attrs["all_times_parameter_assign"]
    all_times_submit = dataset.attrs["all_times_submit"]
    all_times_retrieve = dataset.attrs["all_times_retrieve"]

    transpiled_qc_list = []
    for _, value in dataset.attrs["transpiled_circuits"].items():
        for _, transpiled_circuit in value.items():
            transpiled_qc_list.extend(transpiled_circuit)

    # CLOPS_V
    clops_v: float = num_circuits * num_updates * num_shots * depth / clops_time

    # CLOPS_H
    clops_h: float = 0.0
    counts_depths = {}
    time_count_layers: float = 0.0
    if clops_h_bool:
        # CLOPS_h: count the number of layers made of parallel 2Q gates in the transpiled circuits
        clops_h_depths: List[int]
        qcvv_logger.info("Counting the number of parallel 2Q layer depths in each circuit")
        clops_h_depths, time_count_layers = count_2q_layers(  # pylint: disable=unbalanced-tuple-unpacking
            transpiled_qc_list
        )
        counts_depths = {f"depth_{u}": clops_h_depths.count(u) for u in sorted(set(clops_h_depths))}

        clops_h = num_circuits * num_updates * num_shots * np.mean(clops_h_depths) / clops_time

    processed_results = {
        "clops_v": {"value": int(clops_v), "uncertainty": np.NaN},
        "clops_h": {"value": int(clops_h), "uncertainty": np.NaN},
    }

    dataset.attrs.update(
        {
            "assign_parameters_total": sum(all_times_parameter_assign.values()),
            "user_submit_total": sum(all_times_submit.values()),
            "user_retrieve_total": sum(all_times_retrieve.values()),
            "parallel_2q_layers": counts_depths,
            "time_count_layers": time_count_layers,
        }
    )

    # UPDATE OBSERVATIONS
    observations.update({1: processed_results})

    # PLOT
    # Get all execution elapsed times
    overall_elapsed = retrieve_clops_elapsed_times(all_job_meta)  # will be {} if backend is a simulator
    if overall_elapsed:
        qcvv_logger.info("Total elapsed times from job execution metadata:")
        for k in overall_elapsed.keys():
            dataset.attrs[k] = overall_elapsed[k]
            if overall_elapsed[k] > 60.0:
                qcvv_logger.info(f'\t"{k}": {overall_elapsed[k] / 60.0:.2f} min')
            else:
                qcvv_logger.info(f'\t"{k}": {overall_elapsed[k]:.2f} sec')

        fig_name, fig = plot_times(dataset, observations)
        plots[fig_name] = fig
    else:
        qcvv_logger.info("There is no elapsed-time data associated to jobs (e.g., execution on simulator)")

    # Sort the final dataset
    dataset.attrs = dict(sorted(dataset.attrs.items()))

    return BenchmarkAnalysisResult(dataset=dataset, plots=plots, observations=observations)


class CLOPSBenchmark(Benchmark):
    """
    CLOPS reflect the speed of execution of parametrized QV circuits.
    """

    analysis_function = staticmethod(clops_analysis)

    name: str = "clops"

    def __init__(self, backend_arg: IQMBackendBase | str, configuration: "CLOPSConfiguration"):
        """Construct the QuantumVolumeBenchmark class.

        Args:
            backend_arg (IQMBackendBase | str): the backend to execute the benchmark on
            configuration (QuantumVolumeConfiguration): the configuration of the benchmark
        """
        super().__init__(backend_arg, configuration)

        # EXPERIMENT
        self.backend_configuration_name = backend_arg if isinstance(backend_arg, str) else backend_arg.name
        self.qubits = configuration.qubits
        self.num_qubits = len(self.qubits)
        self.depth = self.num_qubits

        self.u_per_layer = floor(self.num_qubits / 2)
        self.num_parameters = self.depth * self.u_per_layer * 15
        self.param_vector = ParameterVector("p", length=self.num_parameters)

        self.num_circuits = configuration.num_circuits
        self.num_updates = configuration.num_updates
        self.num_shots = configuration.num_shots

        self.qiskit_optim_level = configuration.qiskit_optim_level
        self.optimize_sqg = configuration.optimize_sqg

        # POST-EXPERIMENT AND VARIABLES TO STORE
        self.clops_h_bool = configuration.clops_h_bool

        self.parameters_per_update: Dict[str, List[float]] = {}
        self.counts_per_update: Dict[str, Dict[str, int]] = {}
        self.job_meta_per_update: Dict[str, Dict[str, Dict[str, Any]]] = {}

        self.time_circuit_generate: float = 0.0
        self.time_transpile: float = 0.0
        self.time_sort_batches: float = 0.0

        self.session_timestamp = strftime("%Y-%m-%d_%H:%M:%S")
        self.execution_timestamp: str = ""

    def add_all_meta_to_dataset(self, dataset: xr.Dataset):
        """Adds all configuration metadata and circuits to the dataset variable

        Args:
            dataset (xr.Dataset): The xarray dataset
        """

        dataset.attrs["session_timestamp"] = self.session_timestamp
        dataset.attrs["execution_timestamp"] = self.execution_timestamp
        dataset.attrs["backend_configuration_name"] = self.backend_configuration_name
        dataset.attrs["backend_name"] = self.backend.name

        for key, value in self.configuration:
            if key == "benchmark":  # Avoid saving the class object
                dataset.attrs[key] = value.name
            else:
                dataset.attrs[key] = value

        # Defined outside configuration
        dataset.attrs["num_qubits"] = self.num_qubits
        dataset.attrs["depth"] = self.depth
        dataset.attrs["u_per_layer"] = self.u_per_layer
        dataset.attrs["num_parameters"] = self.num_parameters

    def add_all_circuits_to_dataset(self, dataset: xr.Dataset):
        """Adds all generated circuits during execution to the dataset variable

        Args:
            dataset (xr.Dataset):  The xarray dataset

        Returns:

        """
        qcvv_logger.info(f"Adding all circuits to the dataset")
        for key, circuit in zip(
            ["transpiled_circuits", "untranspiled_circuits"], [self.transpiled_circuits, self.untranspiled_circuits]
        ):
            dictionary = {}
            for outer_key, outer_value in circuit.items():
                dictionary[str(outer_key)] = {
                    str(inner_key): inner_values for inner_key, inner_values in outer_value.items()
                }
            dataset.attrs[key] = dictionary

    def append_parameterized_unitary(
        self,
        qc: QuantumCircuit,
        q0: int,
        q1: int,
        layer: int,
        pair: int,
    ) -> None:
        """Append a decomposed, parametrized SU(4) gate using CX gates to the given quantum circuit.

        Args:
            qc (QuantumCircuit): the quantum circuit to append the SU(4) gate to
            q0 (int): the first qubit involved in the gate
            q1 (int): the second qubit involved in the gate
            layer (int): the QV layer the gate belongs to
            pair (int): the pair index corresponding to the gate
        """
        qc.rz(self.param_vector[15 * (pair + self.u_per_layer * layer) + 0], q0)
        qc.rz(self.param_vector[15 * (pair + self.u_per_layer * layer) + 1], q1)
        qc.sx(q0)
        qc.sx(q1)
        qc.rz(self.param_vector[15 * (pair + self.u_per_layer * layer) + 2], q0)
        qc.rz(self.param_vector[15 * (pair + self.u_per_layer * layer) + 3], q1)
        qc.sx(q0)
        qc.sx(q1)
        qc.rz(self.param_vector[15 * (pair + self.u_per_layer * layer) + 4], q0)
        qc.rz(self.param_vector[15 * (pair + self.u_per_layer * layer) + 5], q1)
        qc.cx(q1, q0)
        qc.rz(self.param_vector[15 * (pair + self.u_per_layer * layer) + 6], q0)
        qc.sx(q1)
        qc.rz(self.param_vector[15 * (pair + self.u_per_layer * layer) + 7], q1)
        qc.cx(q1, q0)
        qc.rz(self.param_vector[15 * (pair + self.u_per_layer * layer) + 8], q0)
        qc.sx(q1)
        qc.cx(q1, q0)
        qc.rz(self.param_vector[15 * (pair + self.u_per_layer * layer) + 9], q0)
        qc.rz(self.param_vector[15 * (pair + self.u_per_layer * layer) + 12], q1)
        qc.sx(q0)
        qc.sx(q1)
        qc.rz(self.param_vector[15 * (pair + self.u_per_layer * layer) + 10], q0)
        qc.rz(self.param_vector[15 * (pair + self.u_per_layer * layer) + 13], q1)
        qc.sx(q0)
        qc.sx(q1)
        qc.rz(self.param_vector[15 * (pair + self.u_per_layer * layer) + 11], q0)
        qc.rz(self.param_vector[15 * (pair + self.u_per_layer * layer) + 14], q1)

    def generate_single_circuit(self) -> QuantumCircuit:
        """Generate a single parametrized QV quantum circuit, with measurements at the end.

        Returns:
            QuantumCircuit: the QV quantum circuit.
        """
        self.u_per_layer = floor(self.num_qubits / 2)
        qc = QuantumCircuit(self.num_qubits)

        qubits = qc.qubits

        for layer in range(self.depth):
            perm = np.random.permutation(self.num_qubits)
            for pair in range(self.u_per_layer):
                q0 = qubits[perm[pair * 2]]
                q1 = qubits[perm[pair * 2 + 1]]

                self.append_parameterized_unitary(qc, q0, q1, layer, pair)

        qc.measure_all()

        return qc

    @timeit
    def generate_circuit_list(
        self,
    ) -> List[QuantumCircuit]:
        """Generate a list of parametrized QV quantum circuits, with measurements at the end.

        Returns:
            List[QuantumCircuit]: the list of parametrized QV quantum circuits.
        """
        qc_list = [self.generate_single_circuit() for _ in range(self.num_circuits)]
        return qc_list

    def generate_random_parameters(self) -> np.ndarray:
        """Generate an array of as many random parameters as needed by the QV template circuits.

        Returns:
            np.ndarray[np.float64]: the array of random parameters
        """
        return np.random.uniform(low=-pi, high=pi, size=self.num_parameters)

    @timeit
    def assign_random_parameters_to_all(
        self,
        dict_parametrized_circs: Dict[Tuple, List[QuantumCircuit]],
        optimize_sqg: bool,
    ) -> Tuple[List[List[float]], Dict[Tuple, List[QuantumCircuit]]]:
        """

        Args:
            dict_parametrized_circs (Dict[Tuple, List[QuantumCircuit]]): Dictionary with list of int (qubits) as keys and lists of parametrized quantum circuits as values
            optimize_sqg (bool): Whether single qubit gate optimization is applied
        Returns:
            A tuple of dictionaries:
            - lists of list of float parameter values corresponding to param updates
            - dictionary with lists of int (qubits) as keys and lists of quantum circuits as values
        """
        # Store parametrized circuits in a separate dictionary
        sorted_dict_parametrized: Dict[Tuple, List[QuantumCircuit]] = {k: [] for k in dict_parametrized_circs.keys()}
        param_values: List[List[float]] = []
        for k in dict_parametrized_circs.keys():
            for qc in dict_parametrized_circs[k]:
                # Update parameters
                parameters = self.generate_random_parameters()

                # NDonis can't use optimize_sqg as is, yet -> complains about MOVE gate not being IQM native!
                if optimize_sqg and self.backend.name != "IQMNdonisBackend":
                    sorted_dict_parametrized[k].append(
                        optimize_single_qubit_gates(  # Optimize SQG seems worth it AFTER assignment
                            qc.assign_parameters(
                                dict(zip(qc.parameters, parameters)),
                                inplace=False,  # Leave the template intact for next updates
                            )
                        )
                    )
                else:
                    sorted_dict_parametrized[k].append(
                        qc.assign_parameters(
                            dict(zip(qc.parameters, parameters)),
                            inplace=False,  # Leave the template intact for next updates
                        )
                    )

                param_values.append(list(parameters))

        return param_values, sorted_dict_parametrized

    def clops_cycle(
        self,
        backend: IQMBackendBase,
        sorted_transpiled_qc_list: Dict[Tuple, List[QuantumCircuit]],
        update: int,
    ) -> Tuple[float, float, float]:
        """Executes a single CLOPS cycle (parameter assignment and execution) for the given update
        Args:
            backend (IQMBackendBase): the backend to execute the jobs with
            sorted_transpiled_qc_list (Dict[str, List[QuantumCircuit]]): A dictionary of lists of transpiled quantum circuits
            update (int): The current cycle update
        Returns:
            Tuple[float, float, float]: The elapsed times for parameter assignment, submission and retrieval of jobs
        """
        # Assign parameters to all quantum circuits
        qcvv_logger.info(
            f"Update {(update + 1)}/{self.num_updates}\nAssigning random parameters to all {self.num_circuits} circuits"
        )
        (all_param_updates, sorted_transpiled_qc_list_parametrized), time_parameter_assign = (
            self.assign_random_parameters_to_all(sorted_transpiled_qc_list, self.optimize_sqg)
        )

        qcvv_logger.info(f"Executing the corresponding circuit batch")
        # Submit all circuits to execute
        all_jobs, time_submit = submit_execute(
            sorted_transpiled_qc_list_parametrized,
            backend,
            self.num_shots,
            self.calset_id,
            max_gates_per_batch=self.max_gates_per_batch,
        )

        qcvv_logger.info(f"Retrieving counts")
        # Retrieve counts - the precise outputs do not matter
        all_counts, time_retrieve = retrieve_all_counts(all_jobs)
        # Save counts - ensures counts were received and can be inspected
        self.parameters_per_update["parameters_update_" + str(update + 1)] = all_param_updates
        self.counts_per_update["counts_update_" + str(update + 1)] = all_counts
        # Retrieve and save all job metadata
        all_job_metadata = retrieve_all_job_metadata(all_jobs)
        self.job_meta_per_update["update_" + str(update + 1)] = all_job_metadata

        return time_parameter_assign, time_submit, time_retrieve

    def generate_transpiled_clops_templates(self) -> Dict[Tuple, List[QuantumCircuit]]:
        """Generates CLOPS circuit templates transpiled to the backend's physical layout

        Returns:
            Dict[str, QuantumCircuit]: a dictionary of quantum circuits with keys being str(qubit layout)
        """
        # Generate the list of QV circuits
        self.untranspiled_circuits = {}
        self.transpiled_circuits = {}

        qc_list, self.time_circuit_generate = self.generate_circuit_list()

        coupling_map = set_coupling_map(self.qubits, self.backend, self.physical_layout)
        # Perform transpilation to backend
        qcvv_logger.info(
            f'Will transpile all {self.num_circuits} circuits according to "{self.physical_layout}" physical layout'
        )

        transpiled_qc_list, self.time_transpile = perform_backend_transpilation(
            qc_list,
            self.backend,
            self.qubits,
            coupling_map=coupling_map,
            qiskit_optim_level=self.qiskit_optim_level,
            optimize_sqg=True,
            routing_method=self.routing_method,
        )

        # Batching
        sorted_transpiled_qc_list = {}
        if self.physical_layout == "fixed":
            sorted_transpiled_qc_list = {tuple(self.qubits): transpiled_qc_list}
        elif self.physical_layout == "batching":
            # Sort circuits according to their final measurement mappings
            (sorted_transpiled_qc_list, _), self.time_sort_batches = sort_batches_by_final_layout(transpiled_qc_list)

        self.untranspiled_circuits.update({str(self.qubits): {str(self.qubits): qc_list}})
        self.transpiled_circuits.update(
            {str(self.qubits): {str(key): value for key, value in sorted_transpiled_qc_list.items()}}
        )
        # self.transpiled_circuits[str(self.qubits)].update(sorted_transpiled_qc_list)

        return sorted_transpiled_qc_list

    def execute(self, backend: IQMBackendBase) -> xr.Dataset:
        """Executes the benchmark"""

        self.execution_timestamp = strftime("%Y-%m-%d_%H:%M:%S")

        dataset = xr.Dataset()
        self.add_all_meta_to_dataset(dataset)

        qcvv_logger.info(
            "NB: CLOPS should be estimated with same qubit layout and optional inputs used to establish QV!"
        )
        if self.num_circuits != 100 or self.num_updates != 10 or self.num_shots != 100:
            qcvv_logger.info(
                f"NB: CLOPS parameters, by definition, are [num_circuits=100, num_updates=10, num_shots=100]"
                f" You chose"
                f" [num_circuits={self.num_circuits}, num_updates={self.num_updates}, num_shots={self.num_shots}]."
            )

        qcvv_logger.info(
            f"Now generating {self.num_circuits} parametrized circuit templates on qubits {self.qubits}",
        )

        sorted_transpiled_qc_list = self.generate_transpiled_clops_templates()

        # *********************************************
        # Start CLOPS timer
        # *********************************************
        start_clops_timer = perf_counter()
        qcvv_logger.info(f"CLOPS time started")
        all_times_parameter_assign = {}
        all_times_submit = {}
        all_times_retrieve = {}
        for n in range(self.num_updates):
            time_parameter_assign, time_submit, time_retrieve = self.clops_cycle(backend, sorted_transpiled_qc_list, n)
            all_times_parameter_assign["update_" + str(n + 1)] = time_parameter_assign
            all_times_submit["update_" + str(n + 1)] = time_submit
            all_times_retrieve["update_" + str(n + 1)] = time_retrieve
        # *********************************************
        # End CLOPS timer
        # *********************************************
        end_clops_timer = perf_counter()

        # COUNT OPERATIONS
        all_op_counts = count_native_gates(backend, [x for y in list(sorted_transpiled_qc_list.values()) for x in y])

        dataset.attrs.update(
            {
                "clops_time": end_clops_timer - start_clops_timer,
                "all_times_parameter_assign": all_times_parameter_assign,
                "all_times_submit": all_times_submit,
                "all_times_retrieve": all_times_retrieve,
                "time_circuit_generate": self.time_circuit_generate,
                "time_transpile": self.time_transpile,
                "time_sort_batches": self.time_sort_batches,
                "parameters_per_update": self.parameters_per_update,
                "job_meta_per_update": self.job_meta_per_update,
                "counts_per_update": self.counts_per_update,
                "operation_counts": all_op_counts,
            }
        )

        self.add_all_circuits_to_dataset(dataset)

        return dataset


class CLOPSConfiguration(BenchmarkConfigurationBase):
    """CLOPS configuration.

    Attributes:
        benchmark (Type[Benchmark]): CLOPS Benchmark.
        qubits (Sequence[int]): The Sequence (List or Tuple) of physical qubit labels in which to run the benchmark.
                            * The physical qubit layout should correspond to the one used to establish QV.
        num_circuits (int): The number of parametrized circuit layouts.
                            * By definition of arXiv:2110.14108 [quant-ph] set to 100.
        num_updates (int): The number of parameter assignment updates per circuit layout to perform.
                            * By definition of arXiv:2110.14108 [quant-ph] set to 10.
        num_shots (int): The number of measurement shots per circuit to perform.
                            * By definition of arXiv:2110.14108 [quant-ph] set to 100.
        clops_h_bool (bool): Whether a CLOPS value with layer definition of CLOPS_H is estimated.
                            * Default is False
                            * This will not estimate a rigorous CLOPS_H value (as loosely defined in www.ibm.com/quantum/blog/quantum-metric-layer-fidelity)
        qiskit_optim_level (int): The Qiskit transpilation optimization level.
                            * The optimization level should correspond to the one used to establish QV.
                            * Default is 3.
        optimize_sqg (bool): Whether Single Qubit Gate Optimization is performed upon transpilation.
                            * The optimize_sqg value should correspond to the one used to establish QV.
                            * Default is True
        routing_method (Literal["basic", "lookahead", "stochastic", "sabre", "none"]): The Qiskit transpilation routing method to use.
                            * The routing_method value should correspond to the one used to establish QV.
                            * Default is "sabre".
        physical_layout (Literal["fixed", "batching"]): Whether the coupling map is restricted to qubits in the input layout or not.
                            - "fixed": Restricts the coupling map to only the specified qubits.
                            - "batching": Considers the full coupling map of the backend and circuit execution is batched per final layout.
                            * Default is "fixed".
    """

    benchmark: Type[Benchmark] = CLOPSBenchmark
    qubits: Sequence[int]
    num_circuits: int = 100
    num_updates: int = 10
    num_shots: int = 100
    clops_h_bool: bool = False
    qiskit_optim_level: int = 3
    optimize_sqg: bool = True



"""
The main algorithm and functions that perform iteration steps
"""

from decimal import Decimal
import sys
import time
from warnings import warn

import numpy as np
import numpy.linalg as la
from scipy.linalg import eig, eigh
from scipy.optimize import minimize
from tqdm import tqdm

from mGST.additional_fns import batch, random_gs, transp
from mGST.low_level_jit import ddA_derivs, ddB_derivs, ddM, dK, dK_dMdM, objf
from mGST.optimization import (
    lineobjf_A_geodesic,
    lineobjf_B_geodesic,
    lineobjf_isom_geodesic,
    tangent_proj,
    update_A_geodesic,
    update_B_geodesic,
    update_K_geodesic,
)
from mGST.reporting.figure_gen import plot_objf


def A_SFN_riem_Hess(K, A, B, y, J, d, r, n_povm, lam=1e-3):
    """Riemannian saddle free Newton step on the POVM parametrization

    Parameters
    ----------
    K : numpy array
        Each subarray along the first axis contains a set of Kraus operators.
        The second axis enumerates Kraus operators for a gate specified by the first axis.
    A : numpy array
        Current POVM parametrization
    B : numpy array
        Current initial state parametrization
    y : numpy array
        2D array of measurement outcomes for sequences in J;
        Each column contains the outcome probabilities for a fixed sequence
    J : numpy array
        2D array where each row contains the gate indices of a gate sequence
    length : int
        Length of the test sequences
    d : int
        Number of different gates in the gate set
    r : int
        Superoperator dimension of the gates given by the square of the physical dimension
    rK : int
        Target Kraus rank
    n_povm : int
        Number of POVM-Elements
    lam : float
        Damping parameter for dampled Newton method; Default: 1e-3

    Returns
    -------
    A_new : numpy array
        Updated POVM parametrization
    """
    pdim = int(np.sqrt(r))
    n = n_povm * pdim
    nt = n_povm * r
    rho = (B @ B.T.conj()).reshape(-1)
    H = np.zeros((2, nt, 2, nt)).astype(np.complex128)
    P_T = np.zeros((2, nt, 2, nt)).astype(np.complex128)
    Fyconjy = np.zeros((n_povm, r, n_povm, r)).astype(np.complex128)
    Fyy = np.zeros((n_povm, r, n_povm, r)).astype(np.complex128)

    X = np.einsum("ijkl,ijnm -> iknlm", K, K.conj()).reshape((d, r, r))
    dA_, dMdM, dMconjdM, dconjdA = ddA_derivs(X, A, B, J, y, r, pdim, n_povm)

    # Second derivatives
    for i in range(n_povm):
        Fyconjy[i, :, i, :] = dMconjdM[i] + dconjdA[i]
        Fyy[i, :, i, :] = dMdM[i]

    # derivative
    Fy = dA_.reshape(n, pdim)
    Y = A.reshape(n, pdim)
    rGrad = Fy.conj() - Y @ Fy.T @ Y
    G = np.array([rGrad, rGrad.conj()]).reshape(-1)

    P = np.eye(n) - Y @ Y.T.conj()
    T = transp(n, pdim)

    # Hessian assembly
    H00 = (
        -(np.kron(Y, Y.T)) @ T @ Fyy.reshape(nt, nt).T
        + Fyconjy.reshape(nt, nt).T.conj()
        - (np.kron(np.eye(n), Y.T @ Fy)) / 2
        - (np.kron(Y @ Fy.T, np.eye(pdim))) / 2
        - (np.kron(P, Fy.T.conj() @ Y.conj())) / 2
    )
    H01 = (
        Fyy.reshape(nt, nt).T.conj()
        - np.kron(Y, Y.T) @ T @ Fyconjy.reshape(nt, nt).T
        + (np.kron(Fy.conj(), Y.T) @ T) / 2
        + (np.kron(Y, Fy.T.conj()) @ T) / 2
    )

    H[0, :, 0, :] = H00
    H[0, :, 1, :] = H01
    H[1, :, 0, :] = H01.conj()
    H[1, :, 1, :] = H00.conj()

    P_T[0, :, 0, :] = np.eye(nt) - np.kron(Y @ Y.T.conj(), np.eye(pdim)) / 2
    P_T[0, :, 1, :] = -np.kron(Y, Y.T) @ T / 2
    P_T[1, :, 0, :] = P_T[0, :, 1, :].conj()
    P_T[1, :, 1, :] = P_T[0, :, 0, :].conj()

    H = H.reshape(2 * nt, 2 * nt) @ P_T.reshape(2 * nt, 2 * nt)

    # saddle free newton method
    H = (H + H.T.conj()) / 2
    evals, U = eigh(H)

    # Disregarding gauge directions (zero eigenvalues of the Hessian)
    # inv_diag = evals.copy()
    # inv_diag[np.abs(evals)<1e-14] = 1
    # inv_diag[np.abs(evals)>1e-14] = np.abs(inv_diag[np.abs(evals)>1e-14]) + lam
    # H_abs_inv = U@np.diag(1/inv_diag)@U.T.conj()

    # Damping all eigenvalues
    H_abs_inv = U @ np.diag(1 / (np.abs(evals) + lam)) @ U.T.conj()

    Delta_A = ((H_abs_inv @ G)[:nt]).reshape(n, pdim)

    Delta = tangent_proj(A, Delta_A, 1, n_povm)[0]

    a = minimize(lineobjf_A_geodesic, 1e-9, args=(Delta, X, A, rho, J, y), method="COBYLA").x
    A_new = update_A_geodesic(A, Delta, a)
    return A_new


def B_SFN_riem_Hess(K, A, B, y, J, d, r, n_povm, lam=1e-3):
    """Riemannian saddle free Newton step on the initial state parametrization

    Parameters
    ----------
    K : numpy array
        Each subarray along the first axis contains a set of Kraus operators.
        The second axis enumerates Kraus operators for a gate specified by the first axis.
    A : numpy array
        Current POVM parametrization
    B : numpy array
        Current initial state parametrization
    y : numpy array
        2D array of measurement outcomes for sequences in J;
        Each column contains the outcome probabilities for a fixed sequence
    J : numpy array
        2D array where each row contains the gate indices of a gate sequence
    length : int
        Length of the test sequences
    d : int
        Number of different gates in the gate set
    r : int
        Superoperator dimension of the gates given by the square of the physical dimension
    rK : int
        Target Kraus rank
    n_povm : int
        Number of POVM-Elements
    lam : float
        Damping parameter for dampled Newton method; Default: 1e-3

    Returns
    -------
    B_new : numpy array
        Updated initial state parametrization
    """

    pdim = int(np.sqrt(r))
    n = r
    nt = r
    E = np.array([(A[i].T.conj() @ A[i]).reshape(-1) for i in range(n_povm)])
    H = np.zeros((2, nt, 2, nt)).astype(np.complex128)
    P_T = np.zeros((2, nt, 2, nt)).astype(np.complex128)
    Fyconjy = np.zeros((r, r)).astype(np.complex128)
    Fyy = np.zeros((r, r)).astype(np.complex128)

    X = np.einsum("ijkl,ijnm -> iknlm", K, K.conj()).reshape((d, r, r))
    dB_, dMdM, dMconjdM, dconjdB = ddB_derivs(X, A, B, J, y, r, pdim)

    # Second derivatives
    Fyconjy = dMconjdM + dconjdB
    Fyy = dMdM

    # derivative
    Fy = dB_.reshape(n)
    Y = B.reshape(n)
    rGrad = Fy.conj() - Y * (Fy.T @ Y)
    G = np.array([rGrad, rGrad.conj()]).reshape(-1)

    P = np.eye(n) - np.outer(Y, Y.T.conj())

    # Hessian assembly
    H00 = (
        -(np.outer(Y, Y.T)) @ Fyy.reshape(nt, nt).T
        + Fyconjy.reshape(nt, nt).T.conj()
        - np.eye(n) * (Y.T @ Fy) / 2
        - np.outer(Y, Fy.T) / 2
        - P * (Fy.T.conj() @ Y.conj()) / 2
    )
    H01 = (
        Fyy.reshape(nt, nt).T.conj()
        - np.outer(Y, Y.T) @ Fyconjy.reshape(nt, nt).T
        + np.outer(Fy.conj(), Y.T) / 2
        + np.outer(Y, Fy.T.conj()) / 2
    )

    H[0, :, 0, :] = H00
    H[0, :, 1, :] = H01
    H[1, :, 0, :] = H01.conj()
    H[1, :, 1, :] = H00.conj()

    P_T[0, :, 0, :] = np.eye(nt) - np.outer(Y, Y.T.conj()) / 2
    P_T[0, :, 1, :] = -np.outer(Y, Y.T) / 2
    P_T[1, :, 0, :] = P_T[0, :, 1, :].conj()
    P_T[1, :, 1, :] = P_T[0, :, 0, :].conj()

    H = H.reshape(2 * nt, 2 * nt) @ P_T.reshape(2 * nt, 2 * nt)

    # saddle free newton method
    H = (H + H.T.conj()) / 2
    evals, U = eigh(H)

    # Disregarding gauge directions (zero eigenvalues of the Hessian)
    # inv_diag = evals.copy()
    # inv_diag[np.abs(evals)<1e-14] = 1
    # inv_diag[np.abs(evals)>1e-14] = np.abs(inv_diag[np.abs(evals)>1e-14]) + lam
    # H_abs_inv = U@np.diag(1/inv_diag)@U.T.conj()

    # Damping all eigenvalues
    H_abs_inv = U @ np.diag(1 / (np.abs(evals) + lam)) @ U.T.conj()

    Delta = (H_abs_inv @ G)[:nt]
    # Projection onto tangent space
    Delta = Delta - Y * (Y.T.conj() @ Delta + Delta.T.conj() @ Y) / 2
    res = minimize(lineobjf_B_geodesic, 1e-9, args=(Delta, X, E, B, J, y), method="COBYLA", options={"maxiter": 20})
    a = res.x

    B_new = update_B_geodesic(B, Delta, a)
    return B_new


def gd(K, E, rho, y, J, d, r, rK, fixed_gates, ls="COBYLA"):
    """Do Riemannian gradient descent optimization step on gates

    Parameters
    ----------
    K : numpy array
        Each subarray along the first axis contains a set of Kraus operators.
        The second axis enumerates Kraus operators for a gate specified by the first axis.
    E : numpy array
        Current POVM estimate
    rho : numpy array
        Current initial state estimate
    y : numpy array
        2D array of measurement outcomes for sequences in J;
        Each column contains the outcome probabilities for a fixed sequence
    J : numpy array
        2D array where each row contains the gate indices of a gate sequence
    length : int
        Length of the test sequences
    d : int
        Number of different gates in the gate set
    r : int
        Superoperator dimension of the gates given by the square of the physical dimension
    rK : int
        Target Kraus rank
    ls : {"COBYLA", ...}
        Line search method, takes "method" arguments of scipy.optimize.minimize

    Returns
    -------
    K_new : numpy array
        Updated Kraus parametrizations

    Notes:
        Gradient descent using the Riemannian gradient and updating along the geodesic.
        The step size is determined by minimizing the objective function in the step size parameter.
    """
    # setup
    pdim = int(np.sqrt(r))
    n = rK * pdim
    Delta = np.zeros((d, n, pdim)).astype(np.complex128)
    X = np.einsum("ijkl,ijnm -> iknlm", K, K.conj()).reshape((d, r, r))

    dK_ = dK(X, K, E, rho, J, y, d, r, rK)
    for k in np.where(~fixed_gates)[0]:
        # derivative
        Fy = dK_[k].reshape(n, pdim)
        Y = K[k].reshape(n, pdim)
        # Riem. gradient taken from conjugate derivative
        rGrad = Fy.conj() - Y @ Fy.T @ Y
        Delta[k] = rGrad

    Delta = tangent_proj(K, Delta, d, rK)
    res = minimize(lineobjf_isom_geodesic, 1e-8, args=(Delta, K, E, rho, J, y), method=ls, options={"maxiter": 200})
    a = res.x
    K_new = update_K_geodesic(K, Delta, a)

    return K_new


def SFN_riem_Hess(K, E, rho, y, J, d, r, rK, lam=1e-3, ls="COBYLA", fixed_gates=None):
    """Riemannian saddle free Newton step on each gate individually

    Parameters
    ----------
    K : numpy array
        Each subarray along the first axis contains a set of Kraus operators.
        The second axis enumerates Kraus operators for a gate specified by the first axis.
    E : numpy array
        Current POVM estimate
    rho : numpy array
        Current initial state estimate
    y : numpy array
        2D array of measurement outcomes for sequences in J;
        Each column contains the outcome probabilities for a fixed sequence
    J : numpy array
        2D array where each row contains the gate indices of a gate sequence
    length : int
        Length of the test sequences
    d : int
        Number of different gates in the gate set
    r : int
        Superoperator dimension of the gates given by the square of the physical dimension
    rK : int
        Target Kraus rank
    lam : float
        Damping parameter for dampled Newton method; Default: 1e-3
    ls : {"COBYLA", ...}
        Line search method, takes "method" arguments of scipy.optimize.minimize
    fixed_gates : List
        List of gate indices which are not optimized over and assumed as fixed

    Returns
    -------
    K_new : numpy array
        Updated Kraus parametrizations
    """
    # setup
    pdim = int(np.sqrt(r))
    n = rK * pdim
    nt = rK * r
    H = np.zeros((2 * nt, 2 * nt)).astype(np.complex128)
    P_T = np.zeros((2 * nt, 2 * nt)).astype(np.complex128)
    Delta_K = np.zeros((d, rK, pdim, pdim)).astype(np.complex128)
    X = np.einsum("ijkl,ijnm -> iknlm", K, K.conj()).reshape((d, r, r))
    if not fixed_gates:
        fixed_gates = []

    # compute derivatives
    dK_, dM10, dM11 = dK_dMdM(X, K, E, rho, J, y, d, r, rK)
    dd, dconjd = ddM(X, K, E, rho, J, y, d, r, rK)

    # Second derivatives
    Fyconjy = dM11.reshape(d, nt, d, nt) + np.einsum("ijklmnop->ikmojlnp", dconjd).reshape((d, nt, d, nt))
    Fyy = dM10.reshape(d, nt, d, nt) + np.einsum("ijklmnop->ikmojlnp", dd).reshape((d, nt, d, nt))

    for k in np.where(~fixed_gates)[0]:
        Fy = dK_[k].reshape(n, pdim)
        Y = K[k].reshape(n, pdim)
        # riemannian gradient, taken from conjugate derivative
        rGrad = Fy.conj() - Y @ Fy.T @ Y
        G = np.array([rGrad, rGrad.conj()]).reshape(-1)

        P = np.eye(n) - Y @ Y.T.conj()
        T = transp(n, pdim)

        # Riemannian Hessian with correction terms
        H00 = (
            -(np.kron(Y, Y.T)) @ T @ Fyy[k, :, k, :].T
            + Fyconjy[k, :, k, :].T.conj()
            - (np.kron(np.eye(n), Y.T @ Fy)) / 2
            - (np.kron(Y @ Fy.T, np.eye(pdim))) / 2
            - (np.kron(P, Fy.T.conj() @ Y.conj())) / 2
        )
        H01 = (
            Fyy[k, :, k, :].T.conj()
            - np.kron(Y, Y.T) @ T @ Fyconjy[k, :, k, :].T
            + (np.kron(Fy.conj(), Y.T) @ T) / 2
            + (np.kron(Y, Fy.T.conj()) @ T) / 2
        )

        H[:nt, :nt] = H00
        H[:nt, nt:] = H01
        H[nt:, :nt] = H[:nt, nt:].conj()
        H[nt:, nt:] = H[:nt, :nt].conj()

        # Tangent space projection
        P_T[:nt, :nt] = np.eye(nt) - np.kron(Y @ Y.T.conj(), np.eye(pdim)) / 2
        P_T[:nt, nt:] = -np.kron(Y, Y.T) @ T / 2
        P_T[nt:, :nt] = P_T[:nt, nt:].conj()
        P_T[nt:, nt:] = P_T[:nt, :nt].conj()

        H = H @ P_T

        # saddle free newton method
        evals, S = eig(H)

        H_abs_inv = S @ np.diag(1 / (np.abs(evals) + lam)) @ la.inv(S)
        Delta_K[k] = ((H_abs_inv @ G)[:nt]).reshape(rK, pdim, pdim)

    Delta = tangent_proj(K, Delta_K, d, rK)

    res = minimize(lineobjf_isom_geodesic, 1e-8, args=(Delta, K, E, rho, J, y), method=ls, options={"maxiter": 200})
    a = res.x
    K_new = update_K_geodesic(K, Delta, a)

    return K_new


def SFN_riem_Hess_full(K, E, rho, y, J, d, r, rK, lam=1e-3, ls="COBYLA"):
    """Riemannian saddle free Newton step on product manifold of all gates

    Parameters
    ----------
    K : numpy array
        Each subarray along the first axis contains a set of Kraus operators.
        The second axis enumerates Kraus operators for a gate specified by the first axis.
    E : numpy array
        Current POVM estimate
    rho : numpy array
        Current initial state estimate
    y : numpy array
        2D array of measurement outcomes for sequences in J;
        Each column contains the outcome probabilities for a fixed sequence
    J : numpy array
        2D array where each row contains the gate indices of a gate sequence
    length : int
        Length of the test sequences
    d : int
        Number of different gates in the gate set
    r : int
        Superoperator dimension of the gates given by the square of the physical dimension
    rK : int
        Target Kraus rank
    lam : float
        Damping parameter for dampled Newton method; Default: 1e-3
    ls : {"COBYLA", ...}
        Line search method, takes "method" arguments of scipy.optimize.minimize

    Returns
    -------
    K_new : numpy array
        Updated Kraus parametrizations
    """
    pdim = int(np.sqrt(r))
    n = rK * pdim
    nt = rK * r
    H = np.zeros((2, d, nt, 2, d, nt)).astype(np.complex128)
    P_T = np.zeros((2, d, nt, 2, d, nt)).astype(np.complex128)
    G = np.zeros((2, d, nt)).astype(np.complex128)
    X = np.einsum("ijkl,ijnm -> iknlm", K, K.conj()).reshape((d, r, r))

    # compute derivatives
    dK_, dM10, dM11 = dK_dMdM(X, K, E, rho, J, y, d, r, rK)
    dd, dconjd = ddM(X, K, E, rho, J, y, d, r, rK)

    # Second derivatives
    Fyconjy = dM11.reshape(d, nt, d, nt) + np.einsum("ijklmnop->ikmojlnp", dconjd).reshape((d, nt, d, nt))
    Fyy = dM10.reshape(d, nt, d, nt) + np.einsum("ijklmnop->ikmojlnp", dd).reshape((d, nt, d, nt))

    for k in range(d):
        Fy = dK_[k].reshape((n, pdim))
        Y = K[k].reshape((n, pdim))
        rGrad = Fy.conj() - Y @ Fy.T @ Y

        G[0, k, :] = rGrad.reshape(-1)
        G[1, k, :] = rGrad.conj().reshape(-1)

        P = np.eye(n) - Y @ Y.T.conj()
        T = transp(n, pdim)
        H00 = (
            -(np.kron(Y, Y.T)) @ T @ Fyy[k, :, k, :].T
            + Fyconjy[k, :, k, :].T.conj()
            - (np.kron(np.eye(n), Y.T @ Fy)) / 2
            - (np.kron(Y @ Fy.T, np.eye(pdim))) / 2
            - (np.kron(P, Fy.T.conj() @ Y.conj())) / 2
        )
        H01 = (
            Fyy[k, :, k, :].T.conj()
            - np.kron(Y, Y.T) @ T @ Fyconjy[k, :, k, :].T
            + (np.kron(Fy.conj(), Y.T) @ T) / 2
            + (np.kron(Y, Fy.T.conj()) @ T) / 2
        )

        # Riemannian Hessian with correction terms
        H[0, k, :, 0, k, :] = H00
        H[0, k, :, 1, k, :] = H01
        H[1, k, :, 0, k, :] = H01.conj()
        H[1, k, :, 1, k, :] = H00.conj()

        # Tangent space projection
        P_T[0, k, :, 0, k, :] = np.eye(nt) - np.kron(Y @ Y.T.conj(), np.eye(pdim)) / 2
        P_T[0, k, :, 1, k, :] = -np.kron(Y, Y.T) @ T / 2
        P_T[1, k, :, 0, k, :] = P_T[0, k, :, 1, k, :].conj()
        P_T[1, k, :, 1, k, :] = P_T[0, k, :, 0, k, :].conj()

        for k2 in range(d):
            if k2 != k:
                Yk2 = K[k2].reshape(n, pdim)
                H[0, k2, :, 0, k, :] = Fyconjy[k, :, k2, :].T.conj() - np.kron(Yk2, Yk2.T) @ T @ Fyy[k, :, k2, :].T
                H[0, k2, :, 1, k, :] = Fyy[k, :, k2, :].T.conj() - np.kron(Yk2, Yk2.T) @ T @ Fyconjy[k, :, k2, :].T
                H[1, k2, :, 0, k, :] = H[0, k2, :, 1, k, :].conj()
                H[1, k2, :, 1, k, :] = H[0, k2, :, 0, k, :].conj()

    H = H.reshape(2 * d * nt, -1) @ P_T.reshape((2 * d * nt, -1))

    # application of saddle free newton method
    H = (H + H.T.conj()) / 2
    evals, U = eigh(H)

    # Damping all eigenvalues
    H_abs_inv = U @ np.diag(1 / (np.abs(evals) + lam)) @ U.T.conj()
    Delta_K = ((H_abs_inv @ G.reshape(-1))[: d * nt]).reshape((d, rK, pdim, pdim))

    # Delta_K is already in tangent space but not to sufficient numerical accuracy
    Delta = tangent_proj(K, Delta_K, d, rK)
    res = minimize(lineobjf_isom_geodesic, 1e-8, args=(Delta, K, E, rho, J, y), method=ls, options={"maxiter": 20})
    a = res.x
    K_new = update_K_geodesic(K, Delta, a)
    return K_new


def optimize(y, J, d, r, rK, n_povm, method, K, rho, A, B, fixed_elements):
    """Full gate set optimization update alternating on E, K and rho

    Parameters
    ----------
    y : numpy array
        2D array of measurement outcomes for sequences in J;
        Each column contains the outcome probabilities for a fixed sequence
    J : numpy array
        2D array where each row contains the gate indices of a gate sequence
    d : int
        Number of different gates in the gate set
    r : int
        Superoperator dimension of the gates given by the square of the physical dimension
    rK : int
        Target Kraus rank
    n_povm : int
        Number of POVM-Elements
    method : {"SFN", "GD"}
        Optimization method, Default: "SFN"
    K : numpy array
        Current estimates of Kraus operators
    E : numpy array
        Current POVM estimate
    rho : numpy array
        Current initial state estimate
    A : numpy array
        Current POVM parametrization
    B : numpy array
        Current initial state parametrization

    Returns
    -------
    K_new : numpy array
        Updated estimates of Kraus operators
    X_new : numpy array
        Updated estimates of superoperatos corresponding to K_new
    E_new : numpy array
        Updated POVM estimate
    rho_new : numpy array
        Updated initial state estimate
    A_new : numpy array
        Updated POVM parametrization
    B_new : numpy array
        Updated initial state parametrization
    """
    if "E" in fixed_elements:
        A_new = A
        E_new = np.array([(A_new[i].T.conj() @ A_new[i]).reshape(-1) for i in range(n_povm)])
    else:
        A_new = A_SFN_riem_Hess(K, A, B, y, J, d, r, n_povm)
        E_new = np.array([(A_new[i].T.conj() @ A_new[i]).reshape(-1) for i in range(n_povm)])

    if any(((f"G%i" % i in fixed_elements) for i in range(d))):
        fixed_gates = np.array([(f"G%i" % i in fixed_elements) for i in range(d)])
        if method == "SFN":
            K_new = SFN_riem_Hess(K, E_new, rho, y, J, d, r, rK, lam=1e-3, ls="COBYLA", fixed_gates=fixed_gates)
        else:
            K_new = gd(K, E_new, rho, y, J, d, r, rK, ls="COBYLA", fixed_gates=fixed_gates)
    else:
        if method == "SFN":
            K_new = SFN_riem_Hess_full(K, E_new, rho, y, J, d, r, rK, lam=1e-3, ls="COBYLA")
        else:
            fixed_gates = np.array([(f"G%i" % i in fixed_elements) for i in range(d)])
            K_new = gd(K, E_new, rho, y, J, d, r, rK, fixed_gates=fixed_gates, ls="COBYLA")

    if "rho" in fixed_elements:
        rho_new = rho
        B_new = B
    else:
        B_new = B_SFN_riem_Hess(K_new, A_new, B, y, J, d, r, n_povm, lam=1e-3)
        rho_new = (B_new @ B_new.T.conj()).reshape(-1)
    X_new = np.einsum("ijkl,ijnm -> iknlm", K_new, K_new.conj()).reshape((d, r, r))
    return K_new, X_new, E_new, rho_new, A_new, B_new


def run_mGST(
    *args,
    method="SFN",
    max_inits=10,
    max_iter=200,
    final_iter=70,
    target_rel_prec=1e-4,
    threshold_multiplier=3,
    fixed_elements=None,
    init=None,
    testing=False,
):  # pylint: disable=too-many-branches
    """Main mGST routine

    Parameters
    ----------
    y : numpy array
        2D array of measurement outcomes for sequences in J;
        Each column contains the outcome probabilities for a fixed sequence
    J : numpy array
        2D array where each row contains the gate indices of a gate sequence
    length : int
        Length of the test sequences
    d : int
        Number of different gates in the gate set
    r : int
        Superoperator dimension of the gates given by the square of the physical dimension
    rK : int
        Target Kraus rank
    n_povm : int
        Number of POVM-Elements
    bsize : int
        Size of the batch (number of sequences)
    meas_samples : int
        Number of samples taken per gate sequence to obtain measurement array y
    method : {"SFN", "GD"}
        Optimization method, Default: "SFN"
    max_iter : int
        Maximum number of iterations on batches; Default: 200
    final_iter : int
        Maximum number of iterations on full data set; Default: 70
    target_rel_prec : float
        Target precision relative to stopping value at which the final iteration loop breaks
    init : [ , , ]
        List of 3 numpy arrays in the format [X,E,rho], that can be used as an initialization;
        If no initialization is given a random initialization is used

    Returns
    -------
    K : numpy array
        Updated estimates of Kraus operators
    X : numpy array
        Updated estimates of superoperatos corresponding to K_new
    E : numpy array
        Updated POVM estimate
    rho : numpy array
        Updated initial state estimate
    res_list : list
        Collected objective function values after each iteration
    """
    y, J, _, d, r, rK, n_povm, bsize, meas_samples = args
    t0 = time.time()
    pdim = int(np.sqrt(r))
    # stopping criterion (Factor 3 can be increased if model mismatch is high)
    delta = threshold_multiplier * (1 - y.reshape(-1)) @ y.reshape(-1) / len(J) / n_povm / meas_samples

    if not fixed_elements:
        fixed_elements = []

    if any(((f"G%i" % i in fixed_elements) for i in range(d))) and method == "SFN":
        warn(
            f"The SFN method with fixed gates is currently only implemented via \n"
            f"iterative updates over individual gates and might lead to a slower converges \n"
            f"compared to the default SFN method.",
            stacklevel=2,
        )

    success = False
    print(f"Starting mGST optimization...")

    if init:
        K, E = (init[0], init[1])
        # offset small negative eigenvalues for stability
        rho = init[2] + 1e-14 * np.eye(pdim).reshape(-1)
        max_inits = 0
    else:
        K, _, E, rho = random_gs(d, r, rK, n_povm)

    A = np.array([la.cholesky(E[k].reshape(pdim, pdim) + 1e-14 * np.eye(pdim)).T.conj() for k in range(n_povm)])
    B = la.cholesky(rho.reshape(pdim, pdim))
    X = np.einsum("ijkl,ijnm -> iknlm", K, K.conj()).reshape((d, r, r))
    res_list = [objf(X, E, rho, J, y)]

    for i in range(max_inits):
        for _ in tqdm(range(max_iter), file=sys.stdout):
            yb, Jb = batch(y, J, bsize)
            K, X, E, rho, A, B = optimize(yb, Jb, d, r, rK, n_povm, method, K, rho, A, B, fixed_elements)
            res_list.append(objf(X, E, rho, J, y))
            if res_list[-1] < delta:
                print(f"Batch optimization successful, improving estimate over full data....")
                success = True
                break
        if testing:
            plot_objf(res_list, delta, f"Objective function for batch optimization")
        if success:
            break
        print(f"Run ", i, f"failed, trying new initialization...")

    if not success and max_inits > 0:
        print(f"Success threshold not reached, attempting optimization over full data set...")
    for _ in tqdm(range(final_iter), file=sys.stdout):
        K, X, E, rho, A, B = optimize(y, J, d, r, rK, n_povm, method, K, rho, A, B, fixed_elements)
        res_list.append(objf(X, E, rho, J, y))
        if np.abs(res_list[-2] - res_list[-1]) < delta * target_rel_prec:
            break
    if testing:
        plot_objf(res_list, delta, f"Objective function over batches and full data")
    if success or (res_list[-1] < delta):
        print(f"\t Convergence criterion satisfied")
    else:
        print(f"\t Convergence criterion not satisfied,", f"try increasing max_iter or using new initializations.")
    print(
        f"\t Final objective {Decimal(res_list[-1]):.2e}",
        f"in time {(time.time() - t0):.2f}s",
    )
    return K, X, E, rho, res_list



# Copyright 2024 IQM Benchmarks developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Q-score benchmark
"""

import itertools
from time import strftime
from typing import Callable, Dict, List, Optional, Tuple, Type

from matplotlib.figure import Figure
import matplotlib.pyplot as plt
from networkx import Graph
import networkx as nx
import numpy as np
from qiskit import QuantumCircuit
from scipy.optimize import basinhopping, minimize

from iqm.benchmarks.benchmark import BenchmarkBase, BenchmarkConfigurationBase
from iqm.benchmarks.logging_config import qcvv_logger
from iqm.benchmarks.utils import perform_backend_transpilation, retrieve_all_counts, submit_execute, timeit
from iqm.qiskit_iqm.iqm_backend import IQMBackendBase


class QScoreBenchmark(BenchmarkBase):
    """
    Q-score estimates the size of combinatorial optimization problems a given number of qubits can execute with meaningful results.
    """

    def __init__(self, backend: IQMBackendBase, configuration: "QScoreConfiguration"):
        """Construct the QScoreBenchmark class.

        Args:
            backend (IQMBackendBase): the backend to execute the benchmark on
            configuration (QScoreConfiguration): the configuration of the benchmark
        """
        super().__init__(backend, configuration)

        self.num_instances = configuration.num_instances
        self.num_qaoa_layers = configuration.num_qaoa_layers
        self.min_num_nodes = configuration.min_num_nodes
        self.max_num_nodes = configuration.max_num_nodes
        self.use_virtual_node = configuration.use_virtual_node
        self.use_classically_optimized_angles = configuration.use_classically_optimized_angles
        self.choose_qubits_routine = configuration.choose_qubits_routine
        self.qiskit_optim_level = configuration.qiskit_optim_level
        self.optimize_sqg = configuration.optimize_sqg
        self.timestamp = strftime("%Y%m%d-%H%M%S")
        self.seed = configuration.seed

        self.graph_physical: Graph
        self.virtual_nodes: List[Tuple[int, int]]
        self.node_to_qubit: Dict[int, int]
        self.qubit_to_node: Dict[int, int]

        if self.use_classically_optimized_angles and self.num_qaoa_layers > 1:
            raise ValueError("If the `use_classically_optimized_angles` is chosen, the `num_qaoa_layers` must be 1.")

        if self.use_virtual_node and self.num_qaoa_layers > 1:
            raise ValueError("If the `use_virtual_node` is chosen, the `num_qaoa_layers` must be 1.")

        if self.choose_qubits_routine == "custom":
            self.custom_qubits_array = configuration.custom_qubits_array

    @staticmethod
    def name() -> str:
        return "qscore"

    def generate_maxcut_ansatz(  # pylint: disable=too-many-branches
        self,
        graph: Graph,
        theta: list[float],
    ) -> QuantumCircuit:
        """Generate an ansatz circuit for QAOA MaxCut, with measurements at the end.

        Args:
            graph (networkx graph): the MaxCut problem graph
            theta (list[float]): the variational parameters for QAOA, first gammas then betas

        Returns:
            QuantumCircuit: the QAOA ansatz quantum circuit.
        """
        gamma = theta[: self.num_qaoa_layers]
        beta = theta[self.num_qaoa_layers :]

        num_qubits = self.graph_physical.number_of_nodes()

        if self.graph_physical.number_of_nodes() != graph.number_of_nodes():
            num_qubits = self.graph_physical.number_of_nodes()
            # re-label the nodes to be between 0 and _num_qubits
            self.node_to_qubit = {node: qubit for qubit, node in enumerate(list(self.graph_physical.nodes))}
            self.qubit_to_node = dict(enumerate(list(self.graph_physical.nodes)))
        else:
            num_qubits = graph.number_of_nodes()
            self.node_to_qubit = {node: node for node in list(self.graph_physical.nodes)}  # no relabeling
            self.qubit_to_node = self.node_to_qubit

        # in case the graph is trivial: return empty circuit
        if num_qubits == 0:
            return QuantumCircuit(1)

        qaoa_qc = QuantumCircuit(num_qubits)
        for i in range(0, num_qubits):
            qaoa_qc.h(i)
        for layer in range(self.num_qaoa_layers):
            for edge in self.graph_physical.edges():
                i = self.node_to_qubit[edge[0]]
                j = self.node_to_qubit[edge[1]]
                qaoa_qc.rzz(2 * gamma[layer], i, j)

            # include edges of the virtual node as rz terms
            for vn in self.virtual_nodes:
                for edge in graph.edges(vn[0]):
                    # exclude edges between virtual nodes
                    edges_between_virtual_nodes = list(itertools.combinations([i[0] for i in self.virtual_nodes], 2))
                    if set(edge) not in list(map(set, edges_between_virtual_nodes)):
                        # The value of the fixed node defines the sign of the rz gate
                        sign = 1.0
                        if vn[1] == 1:
                            sign = -1.0
                        qaoa_qc.rz(sign * 2.0 * gamma[layer], self.node_to_qubit[edge[1]])

            for i in range(0, num_qubits):
                qaoa_qc.rx(2 * beta[layer], i)
        qaoa_qc.measure_all()

        return qaoa_qc

    @staticmethod
    def cost_function(x: str, graph: Graph) -> int:
        """Returns the number of cut edges in a graph (with minus sign).

        Args:
            x (str): solution bitstring.
            graph (networkx graph): the MaxCut problem graph.

        Returns:
            obj (float): number of cut edges multiplied by -1.
        """

        obj = 0
        for i, j in graph.edges():
            if x[i] != x[j]:
                obj += 1

        return -1 * obj

    def compute_expectation_value(self, counts: Dict[str, int], graph: Graph) -> float:
        """Computes expectation value based on measurement results.

        Args:
            counts (Dict[str, int]): key as bitstring, val as count
            graph (networkx) graph: the MaxCut problem graph

        Returns:
            avg (float): expectation value of the cut edges for number of counts
        """

        avg = 0
        sum_count = 0
        for bitstring_aux, count in counts.items():
            bitstring_aux_list = list(bitstring_aux)[::-1]  # go from qiskit endianness to networkx endianness

            # map the qubits back to nodes
            bitstring = [""] * (len(bitstring_aux_list) + len(self.virtual_nodes))
            for qubit, node in self.qubit_to_node.items():
                bitstring[node] = bitstring_aux_list[qubit]

            # insert virtual node(s) to bitstring
            for virtual_node in self.virtual_nodes:
                if virtual_node[0] is not None:
                    bitstring[virtual_node[0]] = str(virtual_node[1])

            obj = self.cost_function("".join(bitstring), graph)
            avg += obj * count
            sum_count += count

        return avg / sum_count

    def create_objective_function(self, graph: Graph, qubit_set: List[int]) -> Callable:
        """
        Creates a function that maps the parameters to the parametrized circuit,
        runs it and computes the expectation value.

        Args:
            graph (networkx graph): the MaxCut problem graph.
            qubit_set (List[int]): indices of the used qubits.
        Returns:
            callable: function that gives expectation value of the cut edges from counts sampled from the ansatz
        """

        def objective_function(theta):
            qc = self.generate_maxcut_ansatz(graph, theta)

            if len(qc.count_ops()) == 0:
                counts = {"": 1.0}  # to handle the case of physical graph with no edges

            else:
                coupling_map = self.backend.coupling_map.reduce(qubit_set)
                transpiled_qc_list, _ = perform_backend_transpilation(
                    [qc],
                    backend=self.backend,
                    qubits=qubit_set,
                    coupling_map=coupling_map,
                    qiskit_optim_level=self.qiskit_optim_level,
                    optimize_sqg=self.optimize_sqg,
                    routing_method=self.routing_method,
                )

                sorted_transpiled_qc_list = {tuple(qubit_set): transpiled_qc_list}
                # Execute on the backend
                jobs, _ = submit_execute(
                    sorted_transpiled_qc_list,
                    self.backend,
                    self.shots,
                    self.calset_id,
                    max_gates_per_batch=self.max_gates_per_batch,
                )

                counts = retrieve_all_counts(jobs)[0][0]

            return self.compute_expectation_value(counts, graph)

        return objective_function

    @staticmethod
    def calculate_optimal_angles_for_QAOA_p1(graph: Graph) -> List[float]:
        """
        Calculates the optimal angles for single layer QAOA MaxCut ansatz.

        Args:
            graph (networkx graph): the MaxCut problem graph.

        Returns:
            List[float]: optimal angles gamma and beta.

        """

        def get_Zij_maxcut_p1(edge_ij, gamma, beta):
            """
            Calculates <p1_QAOA | Z_i Z_j | p1_QAOA>, assuming ij is edge of G.
            """
            i, j = edge_ij
            di = graph.degree[i]
            dj = graph.degree[j]

            first = np.cos(2 * gamma) ** (di - 1) + np.cos(2 * gamma) ** (dj - 1)
            first *= 0.5 * np.sin(4 * beta) * np.sin(2 * gamma)

            node_list = list(graph.nodes).copy()
            node_list.remove(i)
            node_list.remove(j)
            f1 = 1
            f2 = 1
            for k in node_list:
                if graph.has_edge(i, k) and graph.has_edge(j, k):  # ijk is triangle
                    f1 *= np.cos(4 * gamma)
                elif graph.has_edge(i, k) or graph.has_edge(j, k):  # ijk is no triangle
                    f1 *= np.cos(2 * gamma)
                    f2 *= np.cos(2 * gamma)
            second = 0.5 * np.sin(2 * beta) ** 2 * (f1 - f2)
            return first - second

        def get_expected_zz_edgedensity(x):
            gamma = x[0]
            beta = x[1]
            # pylint: disable=consider-using-generator
            return sum([get_Zij_maxcut_p1(edge, gamma, beta) for edge in graph.edges]) / graph.number_of_edges()

        bounds = [(0.0, np.pi / 2), (-np.pi / 4, 0.0)]
        x_init = [0.15, -0.28]

        minimizer_kwargs = {"method": "L-BFGS-B", "bounds": bounds}
        res = basinhopping(get_expected_zz_edgedensity, x_init, minimizer_kwargs=minimizer_kwargs, niter=10, T=2)

        return res.x

    def run_QAOA(self, graph: Graph, qubit_set: List[int]) -> float:
        """
        Solves the cut size of MaxCut for a graph using QAOA.
        The result is average value sampled from the optimized ansatz.

        Args:
            graph (networkx graph): the MaxCut problem graph.
            qubit_set (List[int]): indices of the used qubits.

        Returns:
            float: the expectation value of the maximum cut size.

        """

        objective_function = self.create_objective_function(graph, qubit_set)

        if self.use_classically_optimized_angles:
            if self.graph_physical.number_of_edges() != 0:
                opt_angles = self.calculate_optimal_angles_for_QAOA_p1(self.graph_physical)
            else:
                opt_angles = [1.0, 1.0]
            res = minimize(objective_function, opt_angles, method="COBYLA", tol=1e-5, options={"maxiter": 0})
        else:
            # Good initial angles from from Wurtz et.al. "The fixed angle conjecture for QAOA on regular MaxCut graphs." arXiv preprint arXiv:2107.00677 (2021).
            OPTIMAL_INITIAL_ANGLES = {
                "1": [-0.616, 0.393 / 2],
                "2": [-0.488, 0.898 / 2, 0.555 / 2, 0.293 / 2],
                "3": [-0.422, 0.798 / 2, 0.937 / 2, 0.609 / 2, 0.459 / 2, 0.235 / 2],
                "4": [-0.409, 0.781 / 2, 0.988 / 2, 1.156 / 2, 0.600 / 2, 0.434 / 2, 0.297 / 2, 0.159 / 2],
                "5": [-0.36, -0.707, -0.823, -1.005, -1.154, 0.632 / 2, 0.523 / 2, 0.390 / 2, 0.275 / 2, 0.149 / 2],
                "6": [
                    -0.331,
                    -0.645,
                    -0.731,
                    -0.837,
                    -1.009,
                    -1.126,
                    0.636 / 2,
                    0.535 / 2,
                    0.463 / 2,
                    0.360 / 2,
                    0.259 / 2,
                    0.139 / 2,
                ],
                "7": [
                    -0.310,
                    -0.618,
                    -0.690,
                    -0.751,
                    -0.859,
                    -1.020,
                    -1.122,
                    0.648 / 2,
                    0.554 / 2,
                    0.490 / 2,
                    0.445 / 2,
                    0.341 / 2,
                    0.244 / 2,
                    0.131 / 2,
                ],
                "8": [
                    -0.295,
                    -0.587,
                    -0.654,
                    -0.708,
                    -0.765,
                    -0.864,
                    -1.026,
                    -1.116,
                    0.649 / 2,
                    0.555 / 2,
                    0.500 / 2,
                    0.469 / 2,
                    0.420 / 2,
                    0.319 / 2,
                    0.231 / 2,
                    0.123 / 2,
                ],
                "9": [
                    -0.279,
                    -0.566,
                    -0.631,
                    -0.679,
                    -0.726,
                    -0.768,
                    -0.875,
                    -1.037,
                    -1.118,
                    0.654 / 2,
                    0.562 / 2,
                    0.509 / 2,
                    0.487 / 2,
                    0.451 / 2,
                    0.403 / 2,
                    0.305 / 2,
                    0.220 / 2,
                    0.117 / 2,
                ],
                "10": [
                    -0.267,
                    -0.545,
                    -0.610,
                    -0.656,
                    -0.696,
                    -0.729,
                    -0.774,
                    -0.882,
                    -1.044,
                    -1.115,
                    0.656 / 2,
                    0.563 / 2,
                    0.514 / 2,
                    0.496 / 2,
                    0.496 / 2,
                    0.436 / 2,
                    0.388 / 2,
                    0.291 / 2,
                    0.211 / 2,
                    0.112 / 2,
                ],
                "11": [
                    -0.257,
                    -0.528,
                    -0.592,
                    -0.640,
                    -0.677,
                    -0.702,
                    -0.737,
                    -0.775,
                    -0.884,
                    -1.047,
                    -1.115,
                    0.656 / 2,
                    0.563 / 2,
                    0.516 / 2,
                    0.504 / 2,
                    0.482 / 2,
                    0.456 / 2,
                    0.421 / 2,
                    0.371 / 2,
                    0.276 / 2,
                    0.201 / 2,
                    0.107 / 2,
                ],
            }

            theta = OPTIMAL_INITIAL_ANGLES[str(self.num_qaoa_layers)]
            bounds = [(-np.pi, np.pi)] * self.num_qaoa_layers + [(0.0, np.pi)] * self.num_qaoa_layers

            res = minimize(
                objective_function,
                theta,
                bounds=bounds,
                method="COBYLA",
                tol=1e-5,
                options={"maxiter": 300},
            )

        return -res.fun

    @staticmethod
    def is_successful(
        approximation_ratio: float,
    ) -> bool:
        """Check whether a Q-score benchmark returned approximation ratio above beta*, therefore being successful.

        This condition checks that the mean approximation ratio is above the beta* = 0.2 threshold.

        Args:
            approximation_ratio (float): the mean approximation ratio of all problem graphs

        Returns:
            bool: whether the Q-score benchmark was successful
        """
        return bool(approximation_ratio > 0.2)

    @staticmethod
    def choose_qubits_naive(num_qubits: int) -> list[int]:
        """Choose the qubits to execute the circuits on, sequentially starting at qubit 0.

        Args:
            num_qubits (int): the number of qubits to choose.

        Returns:
            list[int]: the list of qubits to execute the circuits on.
        """
        if num_qubits == 2:
            return [0, 2]

        return list(range(num_qubits))

    def choose_qubits_custom(self, num_qubits: int) -> list[int]:
        """Choose the qubits to execute the circuits on, according to elements in custom_qubits_array matching num_qubits number of qubits

        Args:
            num_qubits (int): the number of qubits to choose

        Returns:
            list[int]: the list of qubits to execute the circuits on
        """
        if self.custom_qubits_array is None:
            raise ValueError(
                "If the `choose_qubits_custom` routine is chosen, a `custom_qubits_array` must be specified in `QScoreConfiguration`."
            )
        selected_qubits = [qubit_layout for qubit_layout in self.custom_qubits_array if len(qubit_layout) == num_qubits]
        # User may input more than one num_qubits layouts
        if len(selected_qubits) > 1:
            chosen_qubits = selected_qubits[0]
            self.custom_qubits_array.remove(chosen_qubits)
            # The execute_single_benchmark call must be looped through a COPY of custom_qubits_array
        else:
            chosen_qubits = selected_qubits[0]
        return chosen_qubits

    def plot_approximation_ratios(
        self, list_of_num_nodes: list[int], list_of_cut_sizes: list[list[float]]
    ) -> tuple[str, Figure]:
        """Generate the figure of approximation ratios vs number of nodes,
            including standard deviation and the acceptance threshold.

        Args:
            list_of_num_nodes (list[int]): list of problem graph sizes.
            list_of_cut_sizes (list[list[float]]): the list of lists of maximum average cut sizes of problem graph instances for each problem size.

        Returns:
            str: the name of the figure.
            Figure: the figure.
        """

        LAMBDA = 0.178

        approximation_ratio_lists = {
            num_nodes: (np.array(cut_sizes) - num_nodes * (num_nodes - 1) / 8) / (LAMBDA * num_nodes ** (3 / 2))
            for num_nodes, cut_sizes in zip(list_of_num_nodes, list_of_cut_sizes)
        }

        avg_approximation_ratios = {
            num_nodes: np.mean(approximation_ratios)
            for num_nodes, approximation_ratios in approximation_ratio_lists.items()
        }

        std_of_approximation_ratios = {
            num_nodes: np.std(approximation_ratios) / np.sqrt(len(approximation_ratios) - 1)
            for num_nodes, approximation_ratios in approximation_ratio_lists.items()
        }

        nodes = list(avg_approximation_ratios.keys())
        ratios = list(avg_approximation_ratios.values())
        std = list(std_of_approximation_ratios.values())

        fig = plt.figure()
        ax = plt.axes()

        plt.axhline(0.2, color="red", linestyle="dashed", label="Threshold")
        plt.errorbar(
            nodes, ratios, yerr=std, fmt="-o", capsize=10, markersize=8, color="#759DEB", label="Approximation ratio"
        )

        ax.set_ylabel(r"Q-score ratio $\beta(n)$")
        ax.set_xlabel("Number of nodes $(n)$")
        plt.xticks(range(min(nodes), max(nodes) + 1))
        plt.legend(loc="lower right")
        plt.grid(True)

        if self.use_virtual_node and self.use_classically_optimized_angles:
            title = f"Q-score, {self.num_instances} instances, with virtual node and classically optimized angles\nBackend: {self.backend.name} / {self.timestamp}"
        elif self.use_virtual_node and not self.use_classically_optimized_angles:
            title = f"Q-score, {self.num_instances} instances, with virtual node \nBackend: {self.backend.name} / {self.timestamp}"
        elif not self.use_virtual_node and self.use_classically_optimized_angles:
            title = f"Q-score, {self.num_instances} instances, with classically optimized angles\nBackend: {self.backend.name} / {self.timestamp}"
        else:
            title = f"Q-score, {self.num_instances} instances \nBackend: {self.backend.name} / {self.timestamp}"

        plt.title(
            title,
            fontsize=9,
        )
        fig_name = f"{max(nodes)}_nodes_{self.num_instances}_instances.png"

        # Show plot if verbose is True
        plt.gcf().set_dpi(250)
        plt.show()

        plt.close()

        return fig_name, fig

    def execute_single_benchmark(
        self,
        num_nodes: int,
    ) -> tuple[bool, float, list[float], list[int]]:
        """Execute a single benchmark, for a given number of qubits.

        Args:
            num_nodes (int): number of nodes in the MaxCut problem graphs.

        Returns:
            bool: whether the benchmark was successful.
            float: approximation_ratio.
            list[float]: the list of maximum average cut sizes of problem graph instances.
            list[int]: the set of qubits the Q-score benchmark was executed on.
        """

        cut_sizes: list[float] = []
        seed = self.seed

        for i in range(self.num_instances):
            graph = nx.generators.erdos_renyi_graph(num_nodes, 0.5, seed=seed)
            print(f"graph: {graph}")
            self.graph_physical = graph.copy()
            self.virtual_nodes = []
            if self.use_virtual_node:
                virtual_node, _ = max(
                    graph.degree(), key=lambda x: x[1]
                )  # choose the virtual node as the most connected node
                self.virtual_nodes.append(
                    (virtual_node, 1)
                )  # the second element of the tuple is the value assigned to the virtual node
                self.graph_physical.remove_node(self.virtual_nodes[0][0])
            # See if there are any non-connected nodes if so, remove them also
            # and set them to be opposite value to the possible original virtual node
            for node in self.graph_physical.nodes():
                if self.graph_physical.degree(node) == 0:
                    self.virtual_nodes.append((node, 0))
            for vn in self.virtual_nodes:
                if self.graph_physical.has_node(vn[0]):
                    self.graph_physical.remove_node(vn[0])

            # Graph with no edges has cut size = 0
            if graph.number_of_edges() == 0:
                cut_sizes.append(0)
                seed += 1
                qcvv_logger.info(f"Graph {i+1}/{self.num_instances} had no edges: cut size = 0.")
                continue

            # Choose the qubit layout
            qubit_set = []
            if self.choose_qubits_routine.lower() == "naive":
                qubit_set = self.choose_qubits_naive(num_nodes)
            elif self.choose_qubits_routine.lower() == "custom" or self.choose_qubits_routine.lower() == "mapomatic":
                qubit_set = self.choose_qubits_custom(num_nodes)
            else:
                raise ValueError('choose_qubits_routine must either be "naive" or "custom".')

            # Solve the maximum cut size with QAOA
            cut_sizes.append(self.run_QAOA(graph, qubit_set))
            seed += 1
            qcvv_logger.info(f"Solved the MaxCut on graph {i+1}/{self.num_instances}.")

        average_cut_size = np.mean(cut_sizes) - num_nodes * (num_nodes - 1) / 8
        average_best_cut_size = 0.178 * pow(num_nodes, 3 / 2)
        approximation_ratio = float(average_cut_size / average_best_cut_size)

        self.raw_data[num_nodes] = {
            "qubit_set": qubit_set,
            "cut_sizes": cut_sizes,
        }
        self.results[num_nodes] = {
            "qubit_set": qubit_set,
            "is_successful": str(self.is_successful(approximation_ratio)),
            "approximation_ratio": approximation_ratio,
        }

        # Return whether the single Q-score Benchmark was successful and its mean approximation ratio
        # and cut sizes for all instances.
        return self.is_successful(approximation_ratio), approximation_ratio, cut_sizes, qubit_set

    @timeit
    def execute_full_benchmark(self) -> tuple[int, list[float], list[list[float]]]:
        """Execute the full benchmark, starting with self.min_num_nodes nodes up to failure.

        Returns:
            int: the Q-score of the device.
            list[float]: the list of approximation rations over problem graph instances for each problem size.
            list[list[float]]: the list of lists of maximum average cut sizes of problem graph instances for each problem size.
        """
        qscore = 0

        if self.max_num_nodes is None:
            if self.use_virtual_node:
                max_num_nodes = self.backend.num_qubits + 1
            else:
                max_num_nodes = self.backend.num_qubits
        else:
            max_num_nodes = self.max_num_nodes

        approximation_ratios = []
        list_of_cut_sizes = []

        for num_nodes in range(self.min_num_nodes, max_num_nodes + 1):
            qcvv_logger.info(f"Executing on {self.num_instances} random graphs with {num_nodes} nodes.")
            is_successful, approximation_ratio, cut_sizes = self.execute_single_benchmark(num_nodes)[0:3]
            approximation_ratios.append(approximation_ratio)

            list_of_cut_sizes.append(cut_sizes)
            if is_successful:
                qcvv_logger.info(
                    f"Q-Score = {num_nodes} passed with:\nApproximation ratio (Beta): {approximation_ratio:.4f}; Avg MaxCut size: {np.mean(cut_sizes):.4f}"
                )
                qscore = num_nodes
                continue

            qcvv_logger.info(
                f"Q-Score = {num_nodes} failed with \napproximation ratio (Beta): {approximation_ratio:.4f} < 0.2; Avg MaxCut size: {np.mean(cut_sizes):.4f}"
            )

        self.results["qscore"] = qscore
        fig_name, fig = self.plot_approximation_ratios(
            list(range(self.min_num_nodes, max_num_nodes + 1)), list_of_cut_sizes
        )
        self.figures[fig_name] = fig
        return num_nodes, approximation_ratios, list_of_cut_sizes


class QScoreConfiguration(BenchmarkConfigurationBase):
    """Q-score configuration."""

    benchmark: Type[BenchmarkBase] = QScoreBenchmark
    num_instances: int
    num_qaoa_layers: int = 1
    min_num_nodes: int = 2
    max_num_nodes: Optional[int] = None
    use_virtual_node: bool = True
    use_classically_optimized_angles: bool = True
    choose_qubits_routine: str = "naive"
    min_num_qubits: int = 2  # If choose_qubits_routine is "naive"
    custom_qubits_array: Optional[list[list[int]]] = None
    qiskit_optim_level: int = 3
    optimize_sqg: bool = True
    seed: int = 1



"""Tests for GHZ fidelity estimation using the new base class"""

import numpy as np

from iqm.benchmarks.entanglement.ghz import GHZBenchmark, GHZConfiguration
from iqm.qiskit_iqm.fake_backends.fake_apollo import IQMFakeApollo


backend = IQMFakeApollo()


class TestGHZ:
    def test_layouts(self):
        MINIMAL_GHZ = GHZConfiguration(
            state_generation_routine=f"tree",
            custom_qubits_array=
            [
                [0,1],
                  [0,1,3],
                  [0,1,3,4],
                  # [0,1,2,3,4],
                  # [0,1,2,3,4,5],
                  # [0,1,2,3,4,5,6],
            ],
            shots=3,
            qiskit_optim_level=3,
            optimize_sqg=True,
            fidelity_routine="coherences",
            num_RMs=10,
            rem=False,
            mit_shots=10,
        )
        benchmark = GHZBenchmark(backend, MINIMAL_GHZ)
        benchmark.run()
        benchmark.analyze()

    def test_state_routine(self):
        for gen_routine in [f"tree", f"naive", "log_depth"]:
            MINIMAL_GHZ = GHZConfiguration(
                state_generation_routine=gen_routine,
                custom_qubits_array=[[0,1,2,3]],
                shots=3,
                qiskit_optim_level=3,
                optimize_sqg=True,
                fidelity_routine="coherences",
                num_RMs=10,
                rem=False,
                mit_shots=10,
            )
            benchmark = GHZBenchmark(backend, MINIMAL_GHZ)
            benchmark.run()
            benchmark.analyze()

    def test_rem(self):
        for fidelity_routine in [f"coherences", f"randomized_measurements"]:
            MINIMAL_GHZ = GHZConfiguration(
                state_generation_routine=f"tree",
                custom_qubits_array=[[0,1,2,3]],
                shots=3,
                qiskit_optim_level=3,
                optimize_sqg=True,
                fidelity_routine=fidelity_routine,
                num_RMs=10,
                rem=True,
                mit_shots=10,
            )
            benchmark = GHZBenchmark(backend, MINIMAL_GHZ)
            benchmark.run()
            benchmark.analyze()